# ruff: noqa: SLF001, S101, D100, D103
from __future__ import annotations

import enum
import importlib
from datetime import UTC, datetime
from typing import TypeAlias, TypeVar, Union, cast, get_args, get_origin

import betterproto
import betterproto.casing
from betterproto.lib.std.google import protobuf as betterproto_protobuf
from immutable import Immutable

ReturnType: TypeAlias = (
    Immutable
    | enum.Enum
    | int
    | float
    | str
    | bytes
    | bool
    | None
    | datetime
    | list['ReturnType']
    | set['ReturnType']
    | dict[str, 'ReturnType']
)


class _MissingType: ...


MISSING = _MissingType()
META_FIELD_PREFIX_PACKAGE_NAME = 'meta_field_package_name_'
META_FIELD_PREFIX_PACKAGE_NAME_INDEX = 1000


def get_class(message: betterproto.Message | betterproto.Enum) -> type | None:
    class_name = type(message).__name__
    if isinstance(message, betterproto.Enum):
        unspecified_member = next(iter(type(message).__members__.keys()))
        destination_module_path = (
            unspecified_member[: -len('_UNSPECIFIED')].lower().replace('_dot_', '.')
        )
    elif (
        META_FIELD_PREFIX_PACKAGE_NAME_INDEX
        in type(message)._betterproto.field_name_by_number
    ):
        field_name = type(message)._betterproto.field_name_by_number[
            META_FIELD_PREFIX_PACKAGE_NAME_INDEX
        ]
        if field_name.startswith(META_FIELD_PREFIX_PACKAGE_NAME):
            destination_module_path = field_name[
                len(META_FIELD_PREFIX_PACKAGE_NAME) :
            ].replace('_dot_', '.')
        else:
            return None
    else:
        return None

    destination_module = importlib.import_module(destination_module_path)

    lowercase_keys = {i.lower(): i for i in dir(destination_module)}

    return getattr(destination_module, lowercase_keys.get(class_name.lower(), ''), None)


def reduce_group(message: betterproto.Message) -> betterproto.Message:
    assert len(message._group_current) == 1
    attribute = next(iter(message._group_current.values()))

    return getattr(message, attribute)


T = TypeVar('T', bound=betterproto.Message | betterproto.Enum)


def rebuild_object(  # noqa: C901
    message: betterproto.Message | list[betterproto.Message],
) -> ReturnType:
    if isinstance(message, int | float | str | bytes | bool | None) and not isinstance(
        message,
        betterproto.Enum,
    ):
        return cast('ReturnType', message)

    if isinstance(
        message,
        betterproto_protobuf.DoubleValue
        | betterproto_protobuf.FloatValue
        | betterproto_protobuf.Int64Value
        | betterproto_protobuf.UInt64Value
        | betterproto_protobuf.Int32Value
        | betterproto_protobuf.UInt32Value
        | betterproto_protobuf.BoolValue
        | betterproto_protobuf.StringValue
        | betterproto_protobuf.BytesValue,
    ):
        return message.value

    if isinstance(message, betterproto_protobuf.Empty):
        return None

    if isinstance(message, list):
        return [rebuild_object(item) for item in message]

    if isinstance(message, dict):
        return {
            key: rebuild_object(value)
            for key, value in message.items()
            if value is not None
        }

    if hasattr(message, '_group_current') and len(message._group_current) > 0:
        return rebuild_object(reduce_group(message))

    destination_class = get_class(message)

    if isinstance(message, betterproto.Enum) and message.name:
        if message.name.endswith('UNSPECIFIED'):
            return None
        return getattr(destination_class, message.name)

    keys = message._betterproto.sorted_field_names
    if len(keys) == 1 and keys[0] == 'items':
        items = [rebuild_object(item) for item in getattr(message, 'items', [])]
        if type(message).__name__.endswith('SetType'):
            return set(items)
        return items

    if destination_class is None:
        msg = f'Class not found for {message}'
        raise ValueError(msg)

    if not isinstance(destination_class, type) or not issubclass(
        destination_class,
        Immutable,
    ):
        msg = f'Parsing {message} is not implemented yet'
        raise NotImplementedError(msg)

    fields = {
        betterproto.casing.snake_case(key): get_field_value(
            destination_class,
            message,
            key,
        )
        for key in keys
        if not key.startswith('meta_field_')
    }

    fields = {key: value for key, value in fields.items() if value is not MISSING}

    return destination_class(**fields)


def get_field_value(
    destination_class: type[Immutable],
    message: betterproto.Message,
    key: str,
) -> ReturnType | _MissingType:
    if getattr(message, key) is None:
        # check if destination_class which is a dataclass has a default value for this
        # field
        field_type = destination_class.__dataclass_fields__.get(key)
        origin = get_origin(field_type)
        is_none_accepted = (
            type(None) in get_args(field_type)
            if origin is Union
            else field_type is type(None)
        )

        if not is_none_accepted:
            return MISSING

        return None

    if key.endswith('_timestamp'):
        return datetime.fromtimestamp(
            getattr(message, key),
            tz=UTC,
        )

    return rebuild_object(getattr(message, key))



"""Google Cloud Speech Recognition Engine Implementation."""

import json
import re
import time
from collections.abc import Generator

from abstraction.speech_recognition_mixin import SpeechRecognitionMixin
from typing_extensions import override

from ubo_app.constants import SPEECH_RECOGNITION_FRAME_RATE
from ubo_app.constants.assistant import (
    GOOGLE_CLOUD_SERVICE_ACCOUNT_KEY_PATTERN,
    GOOGLE_CLOUD_SERVICE_ACCOUNT_KEY_SECRET_ID,
)
from ubo_app.engines.google_cloud import GoogleCloudEngine
from ubo_app.store.main import store
from ubo_app.store.services.speech_recognition import (
    SpeechRecognitionEngineName,
    SpeechRecognitionSetSelectedEngineAction,
)
from ubo_app.utils import secrets
from ubo_app.utils.async_ import ToThreadOptions, create_task, to_thread


class GoogleSpeechRecognitionEngine(GoogleCloudEngine, SpeechRecognitionMixin):
    """Google speech recognition engine using Google Cloud Speech-to-Text."""

    def __init__(self) -> None:
        """Initialize the Google speech recognition engine."""
        super().__init__(label='Google')

    @property
    def engine_name(self) -> SpeechRecognitionEngineName:
        """Return the name of the speech recognition engine."""
        return SpeechRecognitionEngineName.GOOGLE

    def _run_loop(self) -> None:
        from google.cloud import speech_v2 as speech
        from google.oauth2 import service_account

        def requests_stream() -> Generator[
            speech.StreamingRecognizeRequest,
            None,
            None,
        ]:
            """Generate audio stream for transcription."""
            yield self._config_request
            while self.should_be_running():
                try:
                    sample = self.input_queue.get_nowait()
                except IndexError:
                    time.sleep(0.05)
                    continue
                else:
                    yield speech.StreamingRecognizeRequest(audio=sample)

        service_account_info_string = secrets.read_secret(
            GOOGLE_CLOUD_SERVICE_ACCOUNT_KEY_SECRET_ID,
        )
        assert service_account_info_string  # noqa: S101
        service_account_info = json.loads(service_account_info_string)
        streaming_config = speech.StreamingRecognitionConfig(
            config=speech.RecognitionConfig(
                explicit_decoding_config=speech.ExplicitDecodingConfig(
                    sample_rate_hertz=SPEECH_RECOGNITION_FRAME_RATE,
                    audio_channel_count=1,
                    encoding=speech.ExplicitDecodingConfig.AudioEncoding.LINEAR16,
                ),
                language_codes=['en-US'],
                model='long',
            ),
        )
        self._config_request = speech.StreamingRecognizeRequest(
            recognizer=f'projects/{service_account_info["project_id"]}/locations/global/recognizers/_',
            streaming_config=streaming_config,
        )

        credentials = service_account.Credentials.from_service_account_info(
            service_account_info,
        )
        client = speech.SpeechClient(credentials=credentials)
        responses = client.streaming_recognize(requests=requests_stream())

        for response in responses:
            for result in response.results:
                if result.is_final and result.alternatives:
                    create_task(self.report(result=result.alternatives[0].transcript))
                    break

    async def _run(self) -> None:
        """Run the Google speech recognition engine."""
        to_thread(
            self._run_loop,
            ToThreadOptions(callback=self._set_task, name='VoskEngine.run'),
        )

    @property
    def is_setup(self) -> bool:
        """Check if the Google speech recognition engine is set up."""
        service_account_info_string = secrets.read_secret(
            GOOGLE_CLOUD_SERVICE_ACCOUNT_KEY_SECRET_ID,
        )
        return (
            bool(service_account_info_string)
            and re.match(
                GOOGLE_CLOUD_SERVICE_ACCOUNT_KEY_PATTERN,
                service_account_info_string,
            )
            is not None
        )

    @override
    async def _setup(self) -> None:
        await super()._setup()
        store.dispatch(
            SpeechRecognitionSetSelectedEngineAction(
                engine_name=self.engine_name,
            ),
        )



"""Ollama assistant engine module."""

from __future__ import annotations

import asyncio
import math

import ollama
from typing_extensions import override

from ubo_app.colors import SUCCESS_COLOR, WARNING_COLOR
from ubo_app.constants.assistant import OLLAMA_SETUP_NOTIFICATION_ID
from ubo_app.engines.abstraction.ai_provider_mixin import AIProviderMixin
from ubo_app.engines.abstraction.needs_setup_mixin import NeedsSetupMixin
from ubo_app.logger import logger
from ubo_app.store.main import store
from ubo_app.store.services.assistant import (
    AssistantLLMName,
    AssistantSetSelectedModelAction,
    AssistantUpdateProvidersAction,
)
from ubo_app.store.services.docker import (
    DockerImageFetchAction,
    DockerImageRunAction,
    DockerItemStatus,
)
from ubo_app.store.services.notifications import (
    Notification,
    NotificationActionItem,
    NotificationDispatchItem,
    NotificationDisplayType,
    NotificationsAddAction,
)
from ubo_app.utils.async_ import create_task
from ubo_app.utils.error_handlers import report_service_error


class OllamaEngine(NeedsSetupMixin, AIProviderMixin):
    """Ollama assistant engine."""

    @property
    def name(self) -> AssistantLLMName:
        """Returns the name identifier for the Ollama assistant."""
        return AssistantLLMName.OLLAMA

    @property
    def label(self) -> str:
        """Returns the display label for the Ollama assistant."""
        return 'Ollama'

    @property
    def not_setup_message(self) -> str:
        """Returns the message shown when Ollama is not set up."""
        return 'Ollama is not set up. Please set it up in the settings.'

    @property
    @override
    @store.with_state(
        lambda state: state.assistant.selected_models[AssistantLLMName.OLLAMA],
    )
    def is_setup(self, model: str) -> bool:  # noqa: PLR0206
        try:
            models = ollama.list().models
        except ConnectionError:
            return False
        else:
            return any(m.model == model for m in models)

    @store.with_state(
        lambda state: state.assistant.selected_models[AssistantLLMName.OLLAMA],
    )
    def _download_model(self, model: str) -> None:
        """Download the specified Ollama model."""

        async def download_ollama_model() -> None:
            """Download Ollama model."""
            client = ollama.AsyncClient()
            progress_notification = Notification(
                id=OLLAMA_SETUP_NOTIFICATION_ID,
                title='Ollama',
                content=f'Downloading {model} model',
                icon='󰇚',
                color=WARNING_COLOR,
                display_type=NotificationDisplayType.STICKY,
                progress=0,
                show_dismiss_action=False,
                dismiss_on_close=False,
                blink=False,
            )
            store.dispatch(NotificationsAddAction(notification=progress_notification))

            try:
                async for response in await client.pull(model, stream=True):
                    store.dispatch(
                        NotificationsAddAction(
                            notification=progress_notification(
                                progress=(response.completed / response.total)
                                if response.completed is not None
                                and response.total is not None
                                else math.nan,
                            ),
                        ),
                    )
            except Exception:
                logger.exception(
                    'Assistant - Error downloading Ollama model',
                    extra={'model': model},
                )
                report_service_error()
            else:
                logger.info(
                    'Ollama model download complete, updating providers',
                    extra={'model': model},
                )
                store.dispatch(
                    NotificationsAddAction(
                        notification=progress_notification(
                            content=f'"{model}" downloaded successfully',
                            icon='󰄬',
                            color=SUCCESS_COLOR,
                            display_type=NotificationDisplayType.FLASH,
                            progress=None,
                        ),
                    ),
                    AssistantSetSelectedModelAction(model=model),
                    AssistantUpdateProvidersAction(),
                )
            finally:
                self.event.set()

        create_task(download_ollama_model())

    @override
    @store.with_state(
        lambda state: state.docker.ollama.status
        if hasattr(state, 'docker')
        else DockerItemStatus.NOT_AVAILABLE,
    )
    async def _setup(self, ollama_status: DockerItemStatus) -> None:
        if self.is_setup:
            return
        self.event = asyncio.Event()
        if ollama_status is DockerItemStatus.RUNNING:
            store.dispatch(
                NotificationsAddAction(
                    notification=Notification(
                        id=OLLAMA_SETUP_NOTIFICATION_ID,
                        title='Ollama Assistant Setup',
                        content='Setting up the Ollama assistant engine.',
                        color=WARNING_COLOR,
                        actions=[
                            NotificationActionItem(
                                label='Download Model',
                                icon='󰇚',
                                action=self._download_model,
                            ),
                        ],
                    ),
                ),
            )
            await self.event.wait()
        elif ollama_status in (
            DockerItemStatus.NOT_AVAILABLE,
            DockerItemStatus.FETCHING,
        ):
            store.dispatch(
                NotificationsAddAction(
                    notification=Notification(
                        id=OLLAMA_SETUP_NOTIFICATION_ID,
                        title='Ollama Assistant Setup',
                        content='The Ollama image is not fetched. It may take a while '
                        'to fetch, try again once download is complete.',
                        color=WARNING_COLOR,
                        actions=[
                            NotificationDispatchItem(
                                label='Download Ollama Image',
                                icon='󰇚',
                                store_action=DockerImageFetchAction(image='ollama'),
                            ),
                        ],
                    ),
                ),
            )
        else:
            store.dispatch(
                NotificationsAddAction(
                    notification=Notification(
                        id=OLLAMA_SETUP_NOTIFICATION_ID,
                        title='Ollama Assistant Setup',
                        content='The Ollama container is not running.',
                        color=WARNING_COLOR,
                        actions=[
                            NotificationDispatchItem(
                                label='Run Ollama Container',
                                icon='󰐊',
                                store_action=DockerImageRunAction(
                                    image='ollama',
                                ),
                            ),
                        ],
                    ),
                ),
            )



"""Update handler in system manager."""

import subprocess
from collections.abc import Generator, Iterator

from ubo_app.constants import UPDATE_ASSETS_PATH
from ubo_app.logger import get_logger

logger = get_logger('system-manager')


def update_handler(target_version: str) -> Iterator[bytes]:
    """Handle update requests."""
    logger.info('Update request received', extra={'target_version': target_version})

    def run_install_script() -> Generator[bytes]:
        process = subprocess.Popen(  # noqa: S603
            [
                (UPDATE_ASSETS_PATH / 'install.sh'),
                *(
                    [
                        f'--target-version={target_version}',
                    ]
                    if target_version is not None
                    else []
                ),
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        if process.stdout is None:
            msg = 'Failed to run install script'
            raise RuntimeError(msg)

        for line in process.stdout:
            logger.debug('Install script output', extra={'line': line})
            if not line:
                stderr = ''
                if process.stderr:
                    stderr = process.stderr.read().decode()

                logger.error(
                    'Failed to update (install script failed)',
                    extra={'stderr': stderr},
                )

                msg = 'Install script failed, no output received.'
                raise RuntimeError(msg)

            yield line

        if process.returncode is None or process.returncode != 0:
            stderr = ''
            if process.stderr:
                stderr = process.stderr.read().decode()

            logger.error(
                'Failed to update (install script failed)',
                extra={'stderr': stderr},
            )

            msg = 'Install script failed, process return code is not 0.'
            raise RuntimeError(msg)

    return run_install_script()



# ruff: noqa: D100, D103
from __future__ import annotations

from dataclasses import replace

from constants import WIFI_STATE_ICON_ID, WIFI_STATE_ICON_PRIORITY, get_signal_icon
from redux import (
    BaseAction,
    CompleteReducerResult,
    InitAction,
    InitializationActionError,
    ReducerResult,
)

from ubo_app.store.services.ethernet import NetState
from ubo_app.store.services.wifi import (
    WiFiAction,
    WiFiEvent,
    WiFiInputConnectionAction,
    WiFiInputConnectionEvent,
    WiFiSetHasVisitedOnboardingAction,
    WiFiState,
    WiFiUpdateAction,
    WiFiUpdateRequestAction,
    WiFiUpdateRequestEvent,
)
from ubo_app.store.status_icons.types import StatusIconsRegisterAction


def reducer(
    state: WiFiState | None,
    action: WiFiAction,
) -> ReducerResult[WiFiState, BaseAction, WiFiEvent]:
    if state is None:
        if isinstance(action, InitAction):
            return CompleteReducerResult(
                state=WiFiState(
                    connections=[],
                    state=NetState.UNKNOWN,
                    current_connection=None,
                ),
                actions=[WiFiUpdateRequestAction()],
            )
        raise InitializationActionError(action)

    match action:
        case WiFiInputConnectionAction():
            return CompleteReducerResult(
                state=state,
                events=[WiFiInputConnectionEvent()],
            )

        case WiFiSetHasVisitedOnboardingAction():
            return CompleteReducerResult(
                state=replace(
                    state,
                    has_visited_onboarding=action.has_visited_onboarding,
                ),
                events=[WiFiUpdateRequestEvent()],
            )

        case WiFiUpdateRequestAction():
            return CompleteReducerResult(
                state=replace(state, connections=None) if action.reset else state,
                events=[WiFiUpdateRequestEvent()],
            )

        case WiFiUpdateAction():
            return CompleteReducerResult(
                state=replace(
                    state,
                    connections=action.connections,
                    state=action.state,
                    current_connection=action.current_connection,
                ),
                actions=[
                    StatusIconsRegisterAction(
                        icon={
                            NetState.CONNECTED: get_signal_icon(
                                action.current_connection.signal_strength
                                if action.current_connection
                                else 0,
                            ),
                            NetState.DISCONNECTED: '󰖪',
                            NetState.PENDING: '󱛇',
                            NetState.NEEDS_ATTENTION: '󱚵',
                            NetState.UNKNOWN: '󰈅',
                        }[action.state],
                        priority=WIFI_STATE_ICON_PRIORITY,
                        id=WIFI_STATE_ICON_ID,
                    ),
                ],
            )

        case _:
            return state

    return state



# ruff: noqa: D100, D103
from __future__ import annotations

import asyncio
import math
import wave
from pathlib import Path
from typing import TYPE_CHECKING, ParamSpec

from audio_manager import AudioManager
from constants import AUDIO_MIC_STATE_ICON_ID, AUDIO_MIC_STATE_ICON_PRIORITY

from ubo_app.colors import DANGER_COLOR, SUCCESS_COLOR, WARNING_COLOR
from ubo_app.store.main import store
from ubo_app.store.services.audio import (
    AudioInstallDriverEvent,
    AudioPlayAudioSampleAction,
    AudioPlayAudioSampleEvent,
    AudioPlayAudioSequenceEvent,
    AudioPlayChimeEvent,
    AudioSample,
)
from ubo_app.store.services.notifications import (
    Chime,
    Notification,
    NotificationDisplayType,
    NotificationsAddAction,
)
from ubo_app.store.status_icons.types import StatusIconsRegisterAction
from ubo_app.utils.async_ import ToThreadOptions, to_thread
from ubo_app.utils.error_handlers import loop_exception_handler
from ubo_app.utils.persistent_store import register_persistent_store
from ubo_app.utils.server import send_command

if TYPE_CHECKING:
    from collections.abc import Callable, Coroutine

    from ubo_app.utils.types import Subscriptions

Args = ParamSpec('Args')


def _run_async_in_thread(
    async_func: Callable[Args, Coroutine],
    *args: Args.args,
    **kwargs: Args.kwargs,
) -> None:
    loop = asyncio.new_event_loop()
    loop.set_exception_handler(loop_exception_handler)
    result = loop.run_until_complete(async_func(*args, **kwargs))
    loop.close()
    return result


async def _install_driver() -> None:
    store.dispatch(
        NotificationsAddAction(
            notification=Notification(
                id='audio_install_driver',
                title='Audio',
                content='Installing driver ...',
                display_type=NotificationDisplayType.STICKY,
                color=WARNING_COLOR,
                icon='󱀞',
                show_dismiss_action=False,
                progress=math.nan,
            ),
        ),
    )
    result = await send_command(
        'audio',
        'install',
        has_output=True,
    )
    if result == 'installed':
        store.dispatch(
            NotificationsAddAction(
                notification=Notification(
                    id='audio_install_driver',
                    title='Audio Driver',
                    content='Installed successfully.\nPlease restart the device.',
                    display_type=NotificationDisplayType.FLASH,
                    color=SUCCESS_COLOR,
                    icon='󰄬',
                    chime=Chime.DONE,
                ),
            ),
        )
    else:
        store.dispatch(
            NotificationsAddAction(
                notification=Notification(
                    id='audio_install_driver',
                    title='Audio Driver',
                    content='Failed to install',
                    display_type=NotificationDisplayType.STICKY,
                    color=DANGER_COLOR,
                    icon='󰜺',
                    chime=Chime.FAILURE,
                ),
            ),
        )


def init_service() -> Subscriptions:
    audio_manager = AudioManager()

    store.dispatch(
        StatusIconsRegisterAction(
            icon='󰍭',
            priority=AUDIO_MIC_STATE_ICON_PRIORITY,
            id=AUDIO_MIC_STATE_ICON_ID,
        ),
    )

    @store.autorun(lambda state: state.audio.playback_volume)
    def set_playback_volume(volume: float) -> None:
        audio_manager.set_playback_volume(volume)

    @store.autorun(lambda state: state.audio.capture_volume)
    def set_capture_valume(volume: float) -> None:
        audio_manager.set_capture_volume(volume)

    @store.autorun(lambda state: state.audio.is_playback_mute)
    def set_playback_mute(is_mute: bool) -> None:  # noqa: FBT001
        audio_manager.set_playback_mute(mute=is_mute)

    _ = set_playback_volume, set_capture_valume, set_playback_mute

    def play_chime(event: AudioPlayChimeEvent) -> None:
        filename = Path(__file__).parent.joinpath(f'sounds/{event.name}.wav').as_posix()
        with wave.open(filename, 'rb') as wave_file:
            sample_rate = wave_file.getframerate()
            channels = wave_file.getnchannels()
            sample_width = wave_file.getsampwidth()
            audio_data = wave_file.readframes(wave_file.getnframes())

            store.dispatch(
                AudioPlayAudioSampleAction(
                    sample=AudioSample(
                        data=audio_data,
                        rate=sample_rate,
                        channels=channels,
                        width=sample_width,
                    ),
                ),
            )

    def play_audio(
        event: AudioPlayAudioSampleEvent | AudioPlayAudioSequenceEvent,
    ) -> None:
        if isinstance(event, AudioPlayAudioSequenceEvent):
            to_thread(
                _run_async_in_thread,
                ToThreadOptions(name='AudioManager.play_sequence'),
                audio_manager.play_sequence,
                event.sample,
                id=event.id,
                index=event.index,
            )
        else:
            to_thread(
                _run_async_in_thread,
                ToThreadOptions(name='AudioManager.play_sample'),
                audio_manager.play_sample,
                event.sample,
            )

    register_persistent_store(
        'audio_state:playback_volume',
        lambda state: state.audio.playback_volume,
    )
    register_persistent_store(
        'audio_state:is_playback_mute',
        lambda state: state.audio.is_playback_mute,
    )
    register_persistent_store(
        'audio_state:capture_volume',
        lambda state: state.audio.capture_volume,
    )
    register_persistent_store(
        'audio_state:is_capture_mute',
        lambda state: state.audio.is_capture_mute,
    )

    return [
        audio_manager.close,
        store.subscribe_event(AudioInstallDriverEvent, _install_driver),
        store.subscribe_event(AudioPlayChimeEvent, play_chime),
        store.subscribe_event(AudioPlayAudioSampleEvent, play_audio),
        store.subscribe_event(AudioPlayAudioSequenceEvent, play_audio),
    ]



"""provides a function to interact with system services."""

from __future__ import annotations

import subprocess
from pathlib import Path

from ubo_app.logger import get_logger


def users_handler(command: str, username: str | None = None) -> str | None:
    """Interact with system services."""
    logger = get_logger('system-manager')
    try:
        if command == 'create':
            result = subprocess.run(  # noqa: S603
                Path(__file__).parent / 'scripts' / 'set_account.sh',
                check=False,
                text=True,
                capture_output=True,
                env={},
            )
            if result.returncode != 0:
                logger.error(
                    'Failed to create account.',
                    extra={'output': result.stdout, 'error': result.stderr},
                )
            result.check_returncode()
            return result.stdout
        if command == 'reset_password' and username:
            result = subprocess.run(  # noqa: S603
                Path(__file__).parent / 'scripts' / 'set_account.sh',
                check=False,
                text=True,
                capture_output=True,
                env={'USERNAME': username},
            )
            if result.returncode != 0:
                logger.error(
                    'Failed to reset password.',
                    extra={
                        'username': username,
                        'output': result.stdout,
                        'error': result.stderr,
                    },
                )
            result.check_returncode()
            return result.stdout
        if command == 'delete' and username:
            result = subprocess.run(  # noqa: S603
                Path(__file__).parent / 'scripts' / 'delete_account.sh',
                check=False,
                text=True,
                capture_output=True,
                env={'USERNAME': username},
            )
            if result.returncode != 0:
                logger.error(
                    'Failed to delete account.',
                    extra={
                        'username': username,
                        'output': result.stdout,
                        'error': result.stderr,
                    },
                )
            result.check_returncode()
    except Exception:
        logger.exception('Failed to handle service command.')
    else:
        return 'done'
    return None



"""Definitions for assistant actions, events and state."""

from __future__ import annotations

import json
from dataclasses import field
from enum import StrEnum
from typing import TYPE_CHECKING, TypeAlias

from immutable import Immutable
from redux import BaseAction, BaseEvent

from ubo_app.constants.assistant import (
    DEFAULT_LLM_CEREBRAS_MODEL,
    DEFAULT_LLM_GOOGLE_MODEL,
    DEFAULT_LLM_GROK_MODEL,
    DEFAULT_LLM_OLLAMA_MODEL,
    DEFAULT_LLM_OLLAMA_ONPREM_MODEL,
    DEFAULT_LLM_OPENAI_MODEL,
)
from ubo_app.utils.persistent_store import read_from_persistent_store

if TYPE_CHECKING:
    from ubo_app.store.services.audio import AudioSample


class AssistantSTTName(StrEnum):
    """Available assistant speech-to-text engines."""

    VOSK = 'vosk'
    GOOGLE_SEGMENTED = 'google_segmented'
    GOOGLE = 'google'
    OPENAI = 'openai'
    DEEPGRAM = 'deepgram'
    ASSEMBLYAI = 'assemblyai'


class AssistantLLMName(StrEnum):
    """Available assistant llms."""

    OLLAMA = 'ollama'
    OLLAMA_ONPREM = 'ollama_onprem'
    GOOGLE = 'google_vertex'
    OPENAI = 'openai'
    GROK = 'grok'
    CEREBRAS = 'cerebras'


DEFAULT_MODELS = {
    AssistantLLMName.OLLAMA: DEFAULT_LLM_OLLAMA_MODEL,
    AssistantLLMName.OLLAMA_ONPREM: DEFAULT_LLM_OLLAMA_ONPREM_MODEL,
    AssistantLLMName.GOOGLE: DEFAULT_LLM_GOOGLE_MODEL,
    AssistantLLMName.OPENAI: DEFAULT_LLM_OPENAI_MODEL,
    AssistantLLMName.GROK: DEFAULT_LLM_GROK_MODEL,
    AssistantLLMName.CEREBRAS: DEFAULT_LLM_CEREBRAS_MODEL,
}


class AssistantTTSName(StrEnum):
    """Available assistant text-to-speech engines."""

    PIPER = 'piper'
    GOOGLE = 'google'
    OPENAI = 'openai'
    ELEVENLABS = 'elevenlabs'
    RIME = 'rime'


class AssistantImageGeneratorName(StrEnum):
    """Available assistant image generator engines."""

    GOOGLE = 'google'
    OPENAI = 'openai'


class McpServerType(StrEnum):
    """MCP server types."""

    STDIO = 'stdio'
    SSE = 'sse'


class McpServerMetadata(Immutable):
    """Metadata for an MCP server."""

    server_id: str  # Format: {name}_{uuid}
    name: str  # User-friendly name
    type: McpServerType  # Server type enum
    config: str  # JSON string for stdio, URL string for sse


class AssistantAction(BaseAction):
    """Base class for assistant actions."""


class AssistantSetIsActiveAction(AssistantAction):
    """Action to set the assistant active state."""

    is_active: bool


class AssistantSetSelectedSTTAction(AssistantAction):
    """Action to set the selected stt."""

    stt_name: AssistantSTTName


class AssistantSetSelectedLLMAction(AssistantAction):
    """Action to set the selected llm."""

    llm_name: AssistantLLMName


class AssistantSetSelectedTTSAction(AssistantAction):
    """Action to set the selected tts."""

    tts_name: AssistantTTSName


class AssistantSetSelectedImageGeneratorAction(AssistantAction):
    """Action to set the selected image generator."""

    image_generator_name: AssistantImageGeneratorName


class AssistantSetSelectedModelAction(AssistantAction):
    """Action to set the selected model."""

    model: str


class AssistantDownloadOllamaModelAction(AssistantAction):
    """Action to download an Ollama model."""

    model: str


class AssistanceFrame(Immutable):
    """An assistance frame."""

    is_last_frame: bool
    timestamp: float
    id: str
    index: int


class AssistanceTextFrame(AssistanceFrame):
    """A text assistance frame."""

    text: str


class AssistanceAudioFrame(AssistanceFrame):
    """An audio assistance frame."""

    audio: AudioSample | None


class AssistanceImageFrame(AssistanceFrame):
    """An image assistance frame."""

    image: bytes
    width: int
    height: int
    format: str
    metadata: dict[str, str]


AcceptableAssistanceFrame: TypeAlias = (
    AssistanceTextFrame | AssistanceAudioFrame | AssistanceImageFrame
)


class AssistantReportAction(AssistantAction):
    """Action to report assistance from the assistant."""

    source_id: str
    data: AcceptableAssistanceFrame


class AssistantStartListeningAction(AssistantAction):
    """Action to start listening for the assistant."""


class AssistantStopListeningAction(AssistantAction):
    """Action to stop listening for the assistant."""


class AssistantToggleListeningAction(AssistantAction):
    """Action to toggle listening state for the assistant."""


class AssistantUpdateProvidersAction(AssistantAction):
    """Action to signal change in the state of available providers."""


class AssistantAddMcpServerAction(AssistantAction):
    """Action to add a new MCP server."""

    name: str
    type: McpServerType
    config: str  # JSON string for stdio, URL string for sse


class AssistantToggleMcpServerAction(AssistantAction):
    """Action to enable/disable an MCP server."""

    server_id: str


class AssistantDeleteMcpServerAction(AssistantAction):
    """Action to delete an MCP server."""

    server_id: str


class AssistantSyncMcpServersAction(AssistantAction):
    """Action to sync MCP servers from filesystem."""


class AssistantEvent(BaseEvent):
    """Base class for assistant events."""


class AssistantDownloadOllamaModelEvent(AssistantEvent):
    """Event to download an Ollama model."""

    model: str


class AssistantHandleReportEvent(AssistantEvent):
    """Action to report assistance from the assistant."""

    source_id: str
    data: AcceptableAssistanceFrame


class AssistantUpdateProvidersEvent(AssistantEvent):
    """Event to signal change in the state of available providers."""


class AssistantAddMcpServerEvent(AssistantEvent):
    """Event to add a new MCP server."""

    name: str
    type: McpServerType
    config: str  # JSON string for stdio, URL string for sse


class AssistantDeleteMcpServerEvent(AssistantEvent):
    """Event to delete an MCP server."""

    server_id: str


class AssistantState(Immutable):
    """State for the assistant service."""

    is_listening: bool = False
    is_microphone_mute: bool = False
    is_active: bool = field(
        default=read_from_persistent_store(
            'assistant:is_active',
            default=False,
        ),
    )
    selected_stt: AssistantSTTName = field(
        default=read_from_persistent_store(
            'assistant:selected_stt',
            default=AssistantSTTName.VOSK,
            mapper=lambda value: AssistantSTTName(value)
            if value in AssistantSTTName.__members__.values()
            else AssistantSTTName.VOSK,
        ),
    )
    selected_llm: AssistantLLMName = field(
        default=read_from_persistent_store(
            'assistant:selected_llm',
            default=AssistantLLMName.OLLAMA,
            mapper=lambda value: AssistantLLMName(value)
            if value in AssistantLLMName.__members__.values()
            else AssistantLLMName.OLLAMA,
        ),
    )
    selected_models: dict[AssistantLLMName, str] = field(
        default_factory=lambda: read_from_persistent_store(
            'assistant:selected_llm_model',
            default=DEFAULT_MODELS,
            mapper=json.loads,
        ),
    )
    selected_tts: AssistantTTSName = field(
        default=read_from_persistent_store(
            'assistant:selected_tts',
            default=AssistantTTSName.PIPER,
            mapper=lambda value: AssistantTTSName(value)
            if value in AssistantTTSName.__members__.values()
            else AssistantTTSName.PIPER,
        ),
    )
    selected_image_generator: AssistantImageGeneratorName = field(
        default=read_from_persistent_store(
            'assistant:selected_image_generator',
            default=AssistantImageGeneratorName.GOOGLE,
            mapper=lambda value: AssistantImageGeneratorName(value)
            if value in AssistantImageGeneratorName.__members__.values()
            else AssistantImageGeneratorName.GOOGLE,
        ),
    )
    mcp_servers: dict[str, McpServerMetadata] = field(default_factory=dict)
    enabled_mcp_servers: list[str] = field(
        default_factory=lambda: read_from_persistent_store(
            'assistant:enabled_mcp_servers',
            default=[],
            mapper=lambda value: json.loads(value)
            if isinstance(value, str)
            else list(value),
        ),
    )
    # JSON string for gRPC autorun - enabled servers with full metadata
    enabled_mcp_servers_with_metadata_json: str = '[]'
    # Setup status for all provider engines - source of truth for UI
    provider_setup_status: dict[str, bool] = field(default_factory=dict)



# ruff: noqa: D100, D101, D102, D103
from __future__ import annotations

import base64
import contextlib
import functools
import inspect
import threading
import weakref
from asyncio import Handle, iscoroutine
from datetime import datetime
from enum import Flag, IntEnum, StrEnum
from types import GenericAlias
from typing import (
    TYPE_CHECKING,
    Any,
    Concatenate,
    Generic,
    Self,
    TypeAlias,
    TypeVar,
    Union,
    cast,
    get_origin,
    overload,
)

import dill
from fake import Fake
from immutable import Immutable, is_immutable
from redux import (
    BaseCombineReducerState,
    CombineReducerAction,
    CombineReducerRegisterAction,
    FinishAction,
    FinishEvent,
    InitAction,
    Store,
    StoreOptions,
    combine_reducers,
)
from redux.autorun import Autorun
from redux.basic_types import (
    Args,
    AutoAwait,
    AutorunOptionsType,
    ReturnType,
    SelectorOutput,
    StrictEvent,
    SubscribeEventCleanup,
)

from ubo_app.constants import STORE_GRACE_PERIOD
from ubo_app.logger import logger
from ubo_app.store.input.reducer import reducer as input_reducer
from ubo_app.store.scheduler import Scheduler
from ubo_app.store.settings.reducer import reducer as settings_reducer
from ubo_app.store.status_icons.reducer import reducer as status_icons_reducer
from ubo_app.store.update_manager.reducer import reducer as update_manager_reducer
from ubo_app.utils.async_ import ToThreadOptions
from ubo_app.utils.error_handlers import report_service_error
from ubo_app.utils.serializer import add_type_field
from ubo_app.utils.service import get_coroutine_runner

if TYPE_CHECKING:
    from collections.abc import Callable, Coroutine

    from redux.basic_types import (
        EventHandler,
        SnapshotAtom,
        TaskCreatorCallback,
    )
    from store.services.file_system import FileSystemAction
    from store.settings.types import SettingsAction

    from ubo_app.store.core.types import MainAction, MainEvent, MainState
    from ubo_app.store.input.types import (
        InputAction,
        InputResolveEvent,
    )
    from ubo_app.store.services.assistant import (
        AssistantAction,
        AssistantEvent,
        AssistantState,
    )
    from ubo_app.store.services.audio import AudioAction, AudioEvent, AudioState
    from ubo_app.store.services.camera import CameraAction, CameraEvent, CameraState
    from ubo_app.store.services.display import DisplayAction, DisplayEvent, DisplayState
    from ubo_app.store.services.docker import DockerAction, DockerState
    from ubo_app.store.services.infrared import (
        InfraredAction,
        InfraredEvent,
        InfraredState,
    )
    from ubo_app.store.services.ip import IpAction, IpEvent, IpState
    from ubo_app.store.services.keypad import KeypadAction, KeypadState
    from ubo_app.store.services.lightdm import LightDMAction, LightDMState
    from ubo_app.store.services.notifications import (
        NotificationsAction,
        NotificationsEvent,
        NotificationsState,
    )
    from ubo_app.store.services.rgb_ring import RgbRingAction, RgbRingState
    from ubo_app.store.services.rpi_connect import RPiConnectAction, RPiConnectState
    from ubo_app.store.services.sensors import SensorsAction, SensorsState
    from ubo_app.store.services.speech_recognition import (
        SpeechRecognitionAction,
        SpeechRecognitionEvent,
        SpeechRecognitionState,
    )
    from ubo_app.store.services.speech_synthesis import (
        SpeechSynthesisAction,
        SpeechSynthesisState,
    )
    from ubo_app.store.services.ssh import SSHAction, SSHState
    from ubo_app.store.services.users import UsersAction, UsersEvent, UsersState
    from ubo_app.store.services.vscode import VSCodeAction, VSCodeState
    from ubo_app.store.services.web_ui import WebUIState
    from ubo_app.store.services.wifi import WiFiAction, WiFiEvent, WiFiState
    from ubo_app.store.settings.types import SettingsState
    from ubo_app.store.status_icons.types import StatusIconsAction, StatusIconsState
    from ubo_app.store.update_manager.types import (
        UpdateManagerAction,
        UpdateManagerState,
    )

UboAction: TypeAlias = Union[
    # Core Actions
    'CombineReducerAction',
    'InitAction',
    'FinishAction',
    'MainAction',
    'SettingsAction',
    'StatusIconsAction',
    'UpdateManagerAction',
    'InputAction',
    # Services Actions
    'AssistantAction',
    'AudioAction',
    'CameraAction',
    'DisplayAction',
    'DockerAction',
    'FileSystemAction',
    'InfraredAction',
    'IpAction',
    'KeypadAction',
    'LightDMAction',
    'NotificationsAction',
    'RgbRingAction',
    'RPiConnectAction',
    'SensorsAction',
    'SpeechRecognitionAction',
    'SpeechSynthesisAction',
    'SSHAction',
    'UsersAction',
    'VSCodeAction',
    'WiFiAction',
]
UboEvent: TypeAlias = Union[
    # Core Events
    'MainEvent',
    'InputResolveEvent',
    # Services Events
    'AssistantEvent',
    'AudioEvent',
    'CameraEvent',
    'DisplayEvent',
    'InfraredEvent',
    'IpEvent',
    'NotificationsEvent',
    'SpeechRecognitionEvent',
    'UsersEvent',
    'WiFiEvent',
]

if threading.current_thread() is not threading.main_thread():
    msg = 'Store should be created in the main thread'
    raise RuntimeError(msg)


class RootState(BaseCombineReducerState):
    main: MainState
    settings: SettingsState
    status_icons: StatusIconsState
    update_manager: UpdateManagerState

    assistant: AssistantState
    audio: AudioState
    camera: CameraState
    display: DisplayState
    docker: DockerState
    infrared: InfraredState
    ip: IpState
    keypad: KeypadState
    lightdm: LightDMState
    notifications: NotificationsState
    rgb_ring: RgbRingState
    rpi_connect: RPiConnectState
    sensors: SensorsState
    speech_recognition: SpeechRecognitionState
    speech_synthesis: SpeechSynthesisState
    ssh: SSHState
    users: UsersState
    vscode: VSCodeState
    web_ui: WebUIState
    wifi: WiFiState


root_reducer, root_reducer_id = combine_reducers(
    state_type=RootState,
    action_type=UboAction,  # pyright: ignore [reportArgumentType]
    event_type=UboEvent,  # pyright: ignore [reportArgumentType]
    settings=settings_reducer,
    status_icons=status_icons_reducer,
    update_manager=update_manager_reducer,
    input=input_reducer,
)

T = TypeVar('T')
LoadedObject = (
    int
    | float
    | str
    | bytes
    | bool
    | None
    | Immutable
    | list['LoadedObject']
    | set['LoadedObject']
)


class _UboEventHandler(Generic[StrictEvent]):
    def __init__(
        self: Self,
        handler: EventHandler[StrictEvent],
        *,
        keep_ref: bool = True,
    ) -> None:
        self.handler_str = str(handler)
        self.handler_name = handler.__name__
        self.handler_qualname = handler.__qualname__
        self.__name__ = f'UboEventHandler:{self.handler_str}'
        self.__qualname__ = f'UboEventHandler:{self.handler_qualname}'
        if keep_ref:
            self.handler_ref = handler
        elif inspect.ismethod(handler):
            self.handler_ref = weakref.WeakMethod(handler)
        else:
            self.handler_ref = weakref.ref(handler)

        self.coroutine_runner = get_coroutine_runner()

    def __call__(self: Self, event: StrictEvent) -> None:
        async def wrapper() -> None:
            if isinstance(self.handler_ref, weakref.ref):
                handler = cast('EventHandler[StrictEvent]', self.handler_ref())
                if not handler:
                    return
            else:
                handler = self.handler_ref

            parameters = 1
            with contextlib.suppress(Exception):
                parameters = len(
                    [
                        param
                        for param in inspect.signature(
                            handler,
                        ).parameters.values()
                        if param.kind
                        in (param.POSITIONAL_ONLY, param.POSITIONAL_OR_KEYWORD)
                    ],
                )

            if parameters == 0:
                result = cast('Callable[[], Any]', handler)()
            else:
                result = cast('Callable[[StrictEvent], Any]', handler)(event)
            if iscoroutine(result):
                await result

        coroutine = wrapper()
        coroutine.__name__ = f'UboEventHandler:wrapped-coroutine:{self.handler_name}'
        coroutine.__qualname__ = (
            f'UboEventHandler:wrapped-coroutine:{self.handler_qualname}'
        )
        from ubo_app.utils.async_ import create_task

        create_task(coroutine, coroutine_runner=self.coroutine_runner)

    def __repr__(self: Self) -> str:
        """Return a string representation of the instance containing the handler."""
        return f'<UboEventHandler:{self.handler_str}>'


class UboStore(Store[RootState, UboAction, UboEvent]):
    @classmethod
    def serialize_value(cls: type[UboStore], obj: object | type) -> SnapshotAtom:  # noqa: C901
        from redux.autorun import Autorun
        from ubo_gui.page import PageWidget

        if isinstance(obj, Autorun):
            obj = obj()
        if isinstance(obj, Flag):
            return obj.value
        if isinstance(obj, set):
            return {'_type': 'set', 'value': [cls.serialize_value(i) for i in obj]}
        if isinstance(obj, bytes):
            return {'_type': 'bytes', 'value': base64.b64encode(obj).decode('utf-8')}
        if isinstance(obj, datetime):
            return {'_type': 'datetime', 'value': obj.isoformat()}
        if isinstance(obj, functools.partial):
            return f'<functools.partial:{cls.serialize_value(obj.func)}>'
        if is_immutable(obj):
            return cls._serialize_dataclass_to_dict(obj)
        if callable(obj):
            return f'<function:{obj.__name__}>'
        if isinstance(obj, dict):
            return {k: cls.serialize_value(v) for k, v in obj.items()}
        if isinstance(obj, Handle | Fake | PageWidget):
            return f'<{type(obj).__name__}>'
        return super().serialize_value(obj)

    @classmethod
    def _serialize_dataclass_to_dict(
        cls: type[UboStore],
        obj: Immutable,
    ) -> dict[str, Any]:
        result = super()._serialize_dataclass_to_dict(obj)
        return add_type_field(obj, result)

    @overload
    def load_object(
        self: Self,
        data: SnapshotAtom,
    ) -> int | float | str | bool | None | Immutable: ...
    @overload
    def load_object(
        self: Self,
        data: SnapshotAtom,
        *,
        object_type: type[T],
    ) -> T: ...

    def load_object(  # noqa: C901, PLR0912
        self: Self,
        data: Any,
        *,
        object_type: GenericAlias | type[T] | None = None,
    ) -> LoadedObject | T:
        if isinstance(data, int | float | str | bool | None):
            return data
        if isinstance(data, list):
            return [self.load_object(i) for i in data]
        if (
            isinstance(data, dict)
            and '_type' in data
            and isinstance(type_ := data.pop('_type'), str)
        ):
            if type_ == 'set':
                return {self.load_object(i) for i in data['value']}
            if type_ == 'bytes':
                return base64.b64decode(data['value'].encode('utf-8'))

            if isinstance(type_, type):
                class_ = type_
            elif isinstance(type_, str):
                class_ = dill.loads(base64.b64decode(type_.encode('utf-8')))  # noqa: S301
            else:
                msg = f'Invalid type {type(type_)}'
                raise TypeError(msg)

            parameters = {key: self.load_object(value) for key, value in data.items()}

            return class_(**parameters)
        if isinstance(object_type, GenericAlias):
            origin = get_origin(object_type)
            if isinstance(data, origin):
                return cast('T', data)
        elif object_type and issubclass(object_type, StrEnum):
            if isinstance(data, str):
                return object_type(data)
            msg = f'Invalid data type {type(data)} for StrEnum {object_type}'
            raise TypeError(msg)
        elif object_type and issubclass(object_type, IntEnum):
            if isinstance(data, int):
                return object_type(data)
            msg = f'Invalid data type {type(data)} for IntEnum {object_type}'
            raise TypeError(msg)
        elif not object_type or isinstance(data, object_type):
            return cast('T', data)

        msg = f'Invalid data type {type(data)}'
        raise TypeError(msg)

    def subscribe_event(
        self: Self,
        event_type: type[StrictEvent],
        handler: EventHandler[StrictEvent],
        *,
        keep_ref: bool = True,
    ) -> SubscribeEventCleanup:
        in_thread_handler = _UboEventHandler(handler, keep_ref=keep_ref)

        if keep_ref:
            return super().subscribe_event(
                event_type,
                in_thread_handler,
                keep_ref=keep_ref,
            )

        # Put the in_thread_handler in the handler's reference island to tie their
        # lifetimes together
        key = f'__ubo_store_in_thread_handler:{event_type.__name__}'

        unsubscribe = super().subscribe_event(
            event_type,
            in_thread_handler,
            keep_ref=keep_ref,
        )

        def unsubscribe_() -> None:
            unsubscribe()
            handlers.remove(in_thread_handler)

        if not hasattr(handler, key):
            handler.__dict__[key] = []

        handlers = handler.__dict__[key]
        handlers.append(in_thread_handler)

        return SubscribeEventCleanup(
            unsubscribe=unsubscribe_,
            handler=in_thread_handler,
        )

    # Maximum actions to process before checking for events
    # Lower = more responsive events, higher = better action throughput
    ACTIONS_PER_EVENT_CHECK = 50

    def run(self: Self) -> None:
        """Override to interleave action and event processing.

        The base _run_actions() has an internal while loop that processes
        ALL queued actions, preventing interleaving. This override inlines
        the action processing logic to process N actions, then ALL pending
        events, then repeat.
        """
        from redux import is_complete_reducer_result, is_state_reducer_result
        from redux.basic_types import FinishAction, FinishEvent

        with self._is_running:
            while len(self._actions) > 0 or len(self._events) > 0:
                # Process a batch of actions (up to ACTIONS_PER_EVENT_CHECK)
                actions_processed = 0
                while (
                    len(self._actions) > 0
                    and actions_processed < self.ACTIONS_PER_EVENT_CHECK
                ):
                    action = self._actions.pop(0)
                    if action is not None:
                        result = self.reducer(self._state, action)
                        if is_complete_reducer_result(result):
                            self._state = result.state
                            if self._state is not None:
                                self._call_listeners(self._state)
                            self._dispatch(
                                [*(result.actions or []), *(result.events or [])],
                            )
                        elif is_state_reducer_result(result):
                            self._state = result
                            if self._state is not None:
                                self._call_listeners(self._state)

                        if isinstance(action, FinishAction):
                            self._dispatch([FinishEvent()])

                    actions_processed += 1

                # Process ALL pending events before continuing with actions
                if len(self._events) > 0:
                    self._run_event_handlers()



CALL_EVENT_KWARGS_KEY = '__ubo_autorun_call_event'


class _UboAutorun(
    Autorun[
        RootState,
        UboAction,
        UboEvent,
        SelectorOutput,
        Any,
        Args,
        ReturnType,
    ],
    Generic[
        SelectorOutput,
        Args,
        ReturnType,
    ],
):
    def __init__(
        self: Self,
        *,
        store: UboStore,
        selector: Callable[[RootState], SelectorOutput],
        comparator: Callable[[RootState], Any] | None,
        func: Callable[
            Concatenate[SelectorOutput, Args],
            ReturnType,
        ],
        options: AutorunOptionsType[ReturnType, AutoAwait],
    ) -> None:
        self.handler_str = str(func)
        self.handler_name = func.__name__
        self.handler_qualname = func.__qualname__
        self.__name__ = f'UboAutorun:{self.handler_str}'
        self.__qualname__ = f'UboAutorun:{self.handler_qualname}'
        if options.keep_ref:
            self.handler_ref = func
        elif inspect.ismethod(func):
            self.handler_ref = weakref.WeakMethod(func)
        else:
            self.handler_ref = weakref.ref(func)

        self.coroutine_runner = get_coroutine_runner()

        super().__init__(
            store=store,
            selector=selector,
            comparator=comparator,
            func=func,
            options=options,
        )

    def call(
        self: Self,
        *args: Args.args,
        **kwargs: Args.kwargs,
    ) -> None:
        call_event: threading.Event | None = cast(
            'threading.Event',
            kwargs.pop(CALL_EVENT_KWARGS_KEY, None),
        )

        def wrapper(super_: Autorun) -> None:
            try:
                super_.call(*args, **kwargs)
            except Exception:
                logger.exception(
                    'Error in autorun call',
                    extra={
                        'autorun': self,
                        'args_': args,
                        'kwargs': kwargs,
                    },
                )
                report_service_error()
            finally:
                if call_event:
                    call_event.set()

        from ubo_app.utils.async_ import to_thread

        to_thread(
            wrapper,
            ToThreadOptions(coroutine_runner=self.coroutine_runner),
            super_=super(),
        )

    def __call__(
        self: Self,
        *args: Args.args,
        **kwargs: Args.kwargs,
    ) -> ReturnType:
        call_event = threading.Event()
        super().__call__(*args, **{**kwargs, CALL_EVENT_KWARGS_KEY: call_event})
        if not call_event.wait(timeout=30):
            report_service_error(
                exception=RuntimeError('Autorun call timed out after 30 seconds'),
            )
        return self._latest_value

    def _create_task(self, coro: Coroutine[None, None, Any]) -> None:
        self.coroutine_runner(coro)


def ubo_create_task(
    coro: Coroutine,
    *,
    callback: TaskCreatorCallback | None = None,
) -> None:
    from ubo_app.utils.async_ import create_task

    create_task(coro, callback)


def action_middleware(action: UboAction) -> UboAction:
    logger.verbose(
        'Action dispatched',
        extra={'action': action},
    )
    return action


def event_middleware(event: UboEvent | FinishEvent) -> UboEvent | FinishEvent | None:
    logger.verbose(
        'Event dispatched',
        extra={'event': event},
    )
    return event


scheduler = Scheduler()
scheduler.start()


store = UboStore(
    root_reducer,
    StoreOptions(
        auto_init=False,
        scheduler=scheduler.set,
        action_middlewares=[action_middleware],
        event_middlewares=[event_middleware],
        task_creator=ubo_create_task,
        on_finish=scheduler.stop,
        grace_time_in_seconds=STORE_GRACE_PERIOD,
        autorun_class=_UboAutorun,
    ),
)


from ubo_app.store.core.reducer import reducer as main_reducer  # noqa: E402

store.dispatch(InitAction())
store.dispatch(
    CombineReducerRegisterAction(
        combine_reducers_id=root_reducer_id,
        key='main',
        reducer=main_reducer,
    ),
)

store._subscribe(  # noqa: SLF001
    lambda state: logger.verbose('State updated', extra={'state': state}),
)



"""Base class for speech recognition engines."""

from __future__ import annotations

import abc

from typing_extensions import override

from ubo_app.engines.abstraction.background_running_mixin import BackgroundRunningMixin
from ubo_app.logger import logger
from ubo_app.store.main import store
from ubo_app.store.services.speech_recognition import (
    SpeechRecognitionSetIsAssistantActiveAction,
    SpeechRecognitionSetIsIntentsActiveAction,
)
from ubo_app.utils.async_evicting_queue import AsyncEvictingQueue


class BaseSpeechRecognitionEngine(BackgroundRunningMixin):
    """Base class for speech recognition engines."""

    @override
    def __init__(self, *, label: str | None = None) -> None:
        """Initialize speech recognition engine."""
        self.input_queue: AsyncEvictingQueue[bytes] = AsyncEvictingQueue(maxsize=5)
        super().__init__(label=label)

    async def queue_audio_chunk(self, chunk: bytes) -> None:
        """Queue a chunk of audio data for processing."""
        await self.input_queue.put(chunk)

    @override
    def run(self) -> bool:
        if not super().run():
            store.dispatch(
                SpeechRecognitionSetIsIntentsActiveAction(is_active=False),
                SpeechRecognitionSetIsAssistantActiveAction(is_active=False),
            )
            return False
        return True

    @abc.abstractmethod
    async def report(self, result: str) -> None:
        """Report the recognized speech."""
        logger.debug(
            'Unprocessed speech recognized',
            extra={
                'result': result,
                'engine_name': self.name,
            },
        )



"""Module for managing the pod ID."""

import pathlib
import random
import string
from typing import Literal, overload

from ubo_app.constants import INSTALLATION_PATH
from ubo_app.utils.eeprom import read_serial_number

id_path = pathlib.Path(INSTALLATION_PATH) / 'pod-id'


@overload
def get_pod_id(*, with_default: Literal[True]) -> str: ...
@overload
def get_pod_id() -> str | None: ...


def get_pod_id(*, with_default: bool = False) -> str | None:
    """Return the pod ID."""
    try:
        return id_path.read_text().strip()
    except FileNotFoundError:
        if with_default:
            return 'ubo-__'
        return None


def set_pod_id() -> None:
    """Set the pod ID."""
    serial_number = read_serial_number()
    available_letters = list(
        set(string.ascii_lowercase + string.digits + '-') - set('I1lO'),
    )

    try:
        random.seed(serial_number)
        # Generate 2 letters random id
        id = f'ubo-{"".join(random.sample(available_letters, 2))}'
        id_path.write_text(id)
    finally:
        random.seed()



# ruff: noqa: D100, D103
from __future__ import annotations

from dataclasses import replace

from redux import InitAction, InitializationActionError

from ubo_app.store.settings.types import SettingsServiceSetStatusAction
from ubo_app.store.status_icons.types import (
    IconState,
    StatusIconsAction,
    StatusIconsRegisterAction,
    StatusIconsState,
)


def reducer(
    state: StatusIconsState | None,
    action: StatusIconsAction | InitAction,
) -> StatusIconsState:
    if state is None:
        if isinstance(action, InitAction):
            return StatusIconsState(icons=[])
        raise InitializationActionError(action)

    match action:
        case StatusIconsRegisterAction() if action.service is not None:
            return replace(
                state,
                icons=sorted(
                    [
                        *[
                            icon_state
                            for icon_state in state.icons
                            if icon_state.id != action.id or icon_state.id is None
                        ],
                        IconState(
                            symbol=action.icon,
                            color=action.color,
                            priority=action.priority,
                            service_id=action.service,
                            id=action.id,
                        ),
                    ],
                    key=lambda entry: entry.priority,
                ),
            )

        case SettingsServiceSetStatusAction() if action.is_active is False:
            return replace(
                state,
                icons=[
                    icon_state
                    for icon_state in state.icons
                    if icon_state.service_id != action.service_id
                ],
            )

        case _:
            return state



"""Docker composition management."""

from __future__ import annotations

import asyncio
import json
import re
import shutil
from dataclasses import replace

from calculate_progress import (
    LayerProgressTracker,
    ServiceTrackers,
    calculate_composition_progress,
)
from docker_images import IMAGES

from ubo_app.colors import DANGER_COLOR
from ubo_app.constants import CONFIG_PATH
from ubo_app.logger import logger
from ubo_app.store.core.types import DeregisterRegularAppAction
from ubo_app.store.main import store
from ubo_app.store.services.docker import (
    DockerImageFetchCompositionEvent,
    DockerImageReleaseCompositionEvent,
    DockerImageRemoveCompositionEvent,
    DockerImageRunCompositionEvent,
    DockerImageSetStatusAction,
    DockerImageStopCompositionEvent,
    DockerItemStatus,
)
from ubo_app.store.services.notifications import (
    Chime,
    Notification,
    NotificationDisplayType,
    NotificationsAddAction,
)
from ubo_app.utils import IS_RPI
from ubo_app.utils.async_ import create_task
from ubo_app.utils.log_process import log_async_process
from ubo_app.utils.server import send_command

COMPOSITIONS_PATH = CONFIG_PATH / 'docker_compositions'
DOCKER_COMPOSITION_FETCH_PROGRESS_NOTIFICATION_ID = (
    'docker:composition_fetch_progress:{}'
)


def _create_composition_fetch_notification(
    composition_id: str,
    label: str,
    content: str,
    progress: float | None = 0,
) -> Notification:
    """Create a fetch progress notification for a composition."""
    return Notification(
        id=DOCKER_COMPOSITION_FETCH_PROGRESS_NOTIFICATION_ID.format(composition_id),
        title=label,
        content=content,
        display_type=NotificationDisplayType.BACKGROUND,
        icon='󰡨',
        show_dismiss_action=False,
        progress=progress,
    )


async def stop_composition(event: DockerImageStopCompositionEvent) -> None:
    """Stop the composition."""
    id = event.image

    store.dispatch(
        DockerImageSetStatusAction(image=id, status=DockerItemStatus.PROCESSING),
    )
    stop_process = await asyncio.subprocess.create_subprocess_exec(
        'docker',
        'compose',
        'stop',
        cwd=COMPOSITIONS_PATH / id,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )
    await stop_process.wait()
    store.dispatch(
        await log_async_process(
            stop_process,
            title='Docker Composition Error',
            message='Failed to stop composition.',
        ),
    )
    await check_composition(id=id)


async def run_composition(event: DockerImageRunCompositionEvent) -> None:
    """Run the composition."""
    id = event.image

    store.dispatch(
        DockerImageSetStatusAction(image=id, status=DockerItemStatus.PROCESSING),
    )
    run_process = await asyncio.subprocess.create_subprocess_exec(
        'docker',
        'compose',
        'up',
        '-d',
        cwd=COMPOSITIONS_PATH / id,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )
    await run_process.wait()
    store.dispatch(
        await log_async_process(
            run_process,
            title='Docker Composition Error',
            message='Failed to run composition.',
        ),
    )
    await check_composition(id=id)


async def remove_composition(event: DockerImageRemoveCompositionEvent) -> None:
    """Remove the composition."""
    id = event.image
    logger.info('Removing composition', extra={'composition_id': id})

    store.dispatch(
        DockerImageSetStatusAction(image=id, status=DockerItemStatus.PROCESSING),
    )

    # Remove composition directory
    # On Pi: use system manager with elevated privileges
    # (Docker creates root-owned directories)
    # On Mac/dev: use shutil.rmtree (no elevated privileges needed)
    try:
        if IS_RPI:
            composition_path = COMPOSITIONS_PATH / id
            await send_command(
                'docker',
                'composition_delete',
                str(composition_path),
            )
        else:
            # Development environment - use shutil directly
            target_path = COMPOSITIONS_PATH / id
            if target_path.exists():
                shutil.rmtree(target_path)
    except Exception:
        logger.exception(
            'Failed to remove composition directory',
            extra={'composition_id': id},
        )

    # Remove from IMAGES dictionary and unregister from Redux
    if id in IMAGES:
        del IMAGES[id]
        logger.info('Removed composition from IMAGES', extra={'composition_id': id})

        # Remove from menu
        store.dispatch(
            DeregisterRegularAppAction(
                key=id,
            ),
        )


async def get_composition_label(composition_id: str) -> str:
    """Get the label for a composition from metadata."""
    try:
        # Try to read from metadata.json
        metadata_path = COMPOSITIONS_PATH / composition_id / 'metadata.json'
        if metadata_path.exists():
            with metadata_path.open() as f:
                metadata = json.load(f)
                return metadata.get('label', composition_id)
    except Exception:
        logger.exception(
            'Failed to read metadata.json',
            extra={'composition_id': composition_id},
        )

    # Fallback to composition_id
    return composition_id.replace('_', ' ').title()


async def _get_composition_image_count(composition_id: str) -> int:
    """Get the number of images in a composition."""
    logger.debug(
        'Starting composition image count',
        extra={
            'composition_id': composition_id,
            'path': COMPOSITIONS_PATH / composition_id,
        },
    )
    try:
        config_process = await asyncio.subprocess.create_subprocess_exec(
            'docker',
            'compose',
            'config',
            '--images',
            cwd=COMPOSITIONS_PATH / composition_id,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        logger.info('Process created, waiting for output',
                    extra={'composition_id': composition_id})
        stdout, stderr = await config_process.communicate()
        logger.info(
            'Got process output',
            extra={
                'composition_id': composition_id,
                'stdout': stdout.decode(),
                'stderr': stderr.decode(),
                'returncode': config_process.returncode,
            },
        )
        image_list = [img for img in stdout.decode().strip().split('\n') if img]
        logger.info(
            'Got composition image list',
            extra={'composition_id': composition_id, 'image_list': image_list},
        )
        return len(image_list)
    except (OSError, ValueError) as e:
        logger.exception(
            'Failed to get composition image count',
            extra={'composition_id': composition_id, 'error': str(e)},
        )
        return 0

def _process_layer_line(
    layer_match: re.Match[str],
    current_service: str | None,
    service_trackers: ServiceTrackers,
) -> None:
    """Process a layer progress line from docker compose output.

    Args:
        layer_match: Regex match object for layer line
        current_service: Name of the current service being processed
        service_trackers: Dict mapping service names to their trackers

    """
    if not current_service:
        return

    layer_id = layer_match.group(1)
    rest_of_line = layer_match.group(2)

    # Check if there's bracketed progress info
    bracket_match = re.search(r'\[(.+?)\]\s*(.+)$', rest_of_line)
    if bracket_match:
        # Format: "status [progress bar] size_info"  # noqa: ERA001
        # Extract status (everything before bracket)
        status = rest_of_line[:rest_of_line.index('[')].strip()
        progress = bracket_match.group(2).strip()  # Size info after bracket
    else:
        # No brackets, entire rest is status
        status = rest_of_line.strip()
        progress = ''

    if current_service in service_trackers:
        service_trackers[current_service].update_layer(
            layer_id,
            status,
            progress,
        )
        logger.debug(
            'Layer update',
            extra={
                'service': current_service,
                'layer': layer_id,
                'status': status,
                'progress': progress,
            },
        )


def _process_service_line(
    service_match: re.Match[str],
    service_trackers: ServiceTrackers,
    completed_services: set[str],
) -> str | None:
    """Process a service-level status line from docker compose output.

    Args:
        service_match: Regex match object for service line
        service_trackers: Dict mapping service names to their trackers
        completed_services: Set to track which services have completed

    Returns:
        The service name if it's currently pulling, None otherwise

    """
    service_name = service_match.group(1)
    status = service_match.group(2)

    # Track which service we're currently processing
    if 'Pulling' in status:
        if service_name not in service_trackers:
            service_trackers[service_name] = LayerProgressTracker(service_name)
        logger.info(
            'Service pulling started',
            extra={'service_name': service_name},
        )
        return service_name

    # Mark service as complete
    if 'Pulled' in status or 'Already exists' in status:
        completed_services.add(service_name)
        # Mark all layers of this service as complete
        if service_name in service_trackers:
            service_trackers[service_name].mark_all_complete()
        logger.info('Service completed', extra={
            'service_name': service_name,
            'status': status,
            'total_completed': len(completed_services),
        })

    return None


async def _stream_pull_output(
    process: asyncio.subprocess.Process,
    service_trackers: ServiceTrackers,
    completed_services: set[str],
) -> None:
    """Stream docker compose pull output and track service progress.

    Args:
        process: The subprocess running docker compose pull
        service_trackers: Dict mapping service names to their LayerProgressTracker
        completed_services: Set to track which services have completed

    """
    if not process.stdout:
        return

    current_service = None

    while True:
        line_bytes = await process.stdout.readline()
        if not line_bytes:
            break

        line = line_bytes.decode('utf-8', errors='ignore').strip()
        if not line:
            continue

        logger.debug('Got line', extra={'line': line})

        # Docker Compose outputs two types of lines:
        # 1. Service-level: "service-name Pulling" or "service-name Pulled"
        # 2. Layer-level: "hash-id status [progress]" (12-char hex hash)

        # Check if this is a layer hash line (12-character hex at start)
        # Format examples:
        # "22f7f53393bf Already exists"
        # "9a80f9a05524 Download complete"
        # "6f22bf4a69c6 Downloading [==> ] 1.5MB/10MB" # noqa: ERA001
        layer_match = re.match(r'^([a-f0-9]{12})\s+(.+)$', line)
        if layer_match:
            _process_layer_line(layer_match, current_service, service_trackers)
        else:
            # Service-level status line
            service_match = re.match(r'^([\w-]+)\s+(.+)', line)
            if service_match:
                new_service = _process_service_line(
                    service_match,
                    service_trackers,
                    completed_services,
                )
                if new_service:
                    current_service = new_service


async def _handle_pull_success(
    base_notification: Notification,
) -> None:
    """Handle successful composition pull."""
    store.dispatch(
        NotificationsAddAction(
            notification=replace(
                base_notification,
                content='All images pulled!',
                progress=1.0,
                display_type=NotificationDisplayType.FLASH,
                show_dismiss_action=True,
                dismiss_on_close=True,
                chime=Chime.DONE,
            ),
        ),
    )


async def _handle_pull_error(
    base_notification: Notification,
    process: asyncio.subprocess.Process | None = None,
) -> None:
    """Handle composition pull error."""
    notification_action = NotificationsAddAction(
        notification=replace(
            base_notification,
            content='Pull failed',
            color=DANGER_COLOR,
            display_type=NotificationDisplayType.FLASH,
            dismiss_on_close=True,
            chime=Chime.FAILURE,
            show_dismiss_action=True,
            progress=None,
        ),
    )

    if process:
        store.dispatch(
            notification_action,
            await log_async_process(
                process,
                title='Docker Composition Error',
                message='Failed to pull composition images.',
            ),
        )
    else:
        store.dispatch(notification_action)


async def pull_composition(event: DockerImageFetchCompositionEvent) -> None:
    """Pull the composition images with progress tracking."""
    from docker_app import prepare_app

    id = event.image
    container = IMAGES[id]

    logger.info('About to call prepare_app', extra={'image': id})
    # Prepare the composition (download files, generate credentials, update metadata)
    if not await prepare_app(container):
        logger.error('prepare_app returned False, exiting', extra={'image': id})
        return

    composition_label = await get_composition_label(id)
    total_images = await _get_composition_image_count(id)

    # Create base notification
    base_notification = _create_composition_fetch_notification(
        id,
        composition_label,
        f'Pulling {total_images} image{"s" if total_images != 1 else ""}...',
    )

    # Dispatch initial status and notification
    store.dispatch(
        DockerImageSetStatusAction(image=id, status=DockerItemStatus.FETCHING),
        NotificationsAddAction(notification=base_notification),
    )

    # Track service progress
    service_trackers: ServiceTrackers = {}
    completed_services: set[str] = set()
    update_interval = 1.0

    try:
        run_process = await asyncio.subprocess.create_subprocess_exec(
            'docker',
            'compose',
            'pull',
            cwd=COMPOSITIONS_PATH / id,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.STDOUT,
        )

        # Stream output and track progress in background
        stream_task = asyncio.create_task(
            _stream_pull_output(run_process, service_trackers, completed_services),
        )

        # Update progress periodically while streaming
        while not stream_task.done():
            # Wait up to update_interval or until streaming completes
            done, _ = await asyncio.wait(
                [stream_task],
                timeout=update_interval,
            )

            # Update progress if task is still running
            if not done:
                progress = calculate_composition_progress(
                    service_trackers,
                    completed_services,
                    total_images,
                )
                # Count completed services
                completed = len(completed_services)
                store.dispatch(
                    NotificationsAddAction(
                        notification=replace(
                            base_notification,
                            content=(
                                f'Pulling images... ({completed}/'
                                f'{total_images if total_images > 0 else "?"}'
                            ),
                            progress=progress,
                        ),
                    ),
                )

        # Ensure streaming task completes
        await stream_task
        await run_process.wait()

        if run_process.returncode != 0:
            await _handle_pull_error(base_notification, run_process)
        else:
            await _handle_pull_success(base_notification)

    except (OSError, ValueError) as e:
        logger.exception(
            'Failed to pull composition',
            extra={'composition_id': id, 'error': str(e)},
        )
        await _handle_pull_error(base_notification)

    finally:
        # Clean up progress trackers to prevent memory leak
        # Each tracker contains layer dictionaries that can grow large
        service_trackers.clear()
        completed_services.clear()
        logger.debug(
            'Cleaned up progress trackers',
            extra={'composition_id': id},
        )
        await check_composition(id=id)


async def _release_composition(id: str) -> None:
    store.dispatch(
        DockerImageSetStatusAction(image=id, status=DockerItemStatus.PROCESSING),
    )
    check_process = await asyncio.subprocess.create_subprocess_exec(
        'docker',
        'compose',
        'down',
        cwd=COMPOSITIONS_PATH / id,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )
    await check_process.wait()
    store.dispatch(
        await log_async_process(
            check_process,
            title='Docker Composition Error',
            message='Failed to release resources.',
        ),
    )
    await check_composition(id=id)


def release_composition(event: DockerImageReleaseCompositionEvent) -> None:
    """Release resources of composition."""
    id = event.image
    create_task(_release_composition(id))


async def check_composition(*, id: str) -> None:
    """Check the status of the composition."""
    # Check if composition directory exists
    composition_path = COMPOSITIONS_PATH / id
    if not composition_path.exists():
        # Directory doesn't exist - set status to NOT_AVAILABLE
        store.dispatch(
            DockerImageSetStatusAction(image=id, status=DockerItemStatus.NOT_AVAILABLE),
        )
        return

    # Check if containers are running
    ps_running = await asyncio.subprocess.create_subprocess_exec(
        'docker',
        'compose',
        'ps',
        '--quiet',
        cwd=composition_path,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )
    # Check all containers (including stopped)
    ps_all = await asyncio.subprocess.create_subprocess_exec(
        'docker',
        'compose',
        'ps',
        '-a',
        '--quiet',
        cwd=composition_path,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )
    # Get required images from compose file
    config = await asyncio.subprocess.create_subprocess_exec(
        'docker',
        'compose',
        'config',
        '--images',
        cwd=composition_path,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )

    await asyncio.gather(
        ps_running.wait(),
        ps_all.wait(),
        config.wait(),
        return_exceptions=True,
    )

    store.dispatch(
        *await asyncio.gather(
            log_async_process(
                ps_running,
                title='Docker Composition Error',
                message='Failed to check running containers.',
            ),
            log_async_process(
                ps_all,
                title='Docker Composition Error',
                message='Failed to check containers.',
            ),
            log_async_process(
                config,
                title='Docker Composition Error',
                message='Failed to check composition config.',
            ),
        ),
    )

    # Check if containers are running
    ps_running_output = await ps_running.stdout.read() if ps_running.stdout else b''
    if ps_running_output:
        store.dispatch(
            DockerImageSetStatusAction(image=id, status=DockerItemStatus.RUNNING),
        )
        return

    # Check if containers exist (even if stopped)
    ps_all_output = await ps_all.stdout.read() if ps_all.stdout else b''
    if ps_all_output:
        store.dispatch(
            DockerImageSetStatusAction(image=id, status=DockerItemStatus.CREATED),
        )
        return

    # Check if images are pulled
    config_output = await config.stdout.read() if config.stdout else b''
    if config_output:
        image_names = config_output.decode().strip().split('\n')
        all_exist = True
        for image_name in image_names:
            if not image_name:
                continue
            inspect = await asyncio.subprocess.create_subprocess_exec(
                'docker',
                'image',
                'inspect',
                image_name,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
            )
            await inspect.wait()
            if inspect.returncode != 0:
                all_exist = False
                break

        status = DockerItemStatus.AVAILABLE if \
                all_exist else DockerItemStatus.NOT_AVAILABLE
        store.dispatch(DockerImageSetStatusAction(image=id, status=status))
    else:
        store.dispatch(
            DockerImageSetStatusAction(image=id, status=DockerItemStatus.NOT_AVAILABLE),
        )



# ruff: noqa: D100
from __future__ import annotations

from typing import TYPE_CHECKING, Protocol

if TYPE_CHECKING:
    import numpy as np
    from numpy.typing import NDArray


class CameraBackend(Protocol):
    """Protocol defining the interface for camera backends."""

    def start(self) -> None:
        """Start the camera."""
        ...

    def stop(self) -> None:
        """Stop the camera."""
        ...

    def close(self) -> None:
        """Release camera resources."""
        ...

    def capture_array(self, stream: str = 'main') -> NDArray[np.uint8] | None:
        """Capture a frame from the camera as a numpy array.

        Args:
            stream: The stream name (for compatibility with PiCamera2)

        Returns:
            Numpy array containing the frame data in RGB format, or None if
            capture fails

        """
        ...

    def configure(self, config: dict | str) -> None:
        """Configure the camera with the given configuration.

        Args:
            config: Configuration dictionary or string

        """
        ...

    def set_controls(self, controls: dict) -> None:
        """Set camera controls.

        Args:
            controls: Dictionary of control parameters

        """
        ...



# ruff: noqa: D100, D101, D102
from __future__ import annotations

import math
from functools import cached_property
from typing import TYPE_CHECKING

from kivy.animation import Animation
from kivy.clock import mainthread
from kivy.metrics import dp
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.label import Label
from kivy.uix.relativelayout import RelativeLayout
from redux import AutorunOptions
from ubo_gui.app import UboApp
from ubo_gui.progress_ring import ProgressRingWidget
from ubo_gui.spinner import SpinnerWidget

from ubo_app.store.main import store

if TYPE_CHECKING:
    from kivy.uix.widget import Widget

    from ubo_app.store.services.notifications import Notification


class MenuAppHeader(UboApp):
    notification_widgets: dict[str, tuple[Notification, Widget]]
    progress_layout: BoxLayout

    @mainthread
    def set_notification_widgets(
        self: MenuAppHeader,
        notifications: list[Notification],
    ) -> None:
        for notification in notifications:
            if notification.progress is None:
                if notification.id in self.notification_widgets:
                    self.progress_layout.remove_widget(
                        self.notification_widgets[notification.id][1],
                    )
                    del self.notification_widgets[notification.id]
            elif math.isnan(notification.progress):
                if notification.id not in self.notification_widgets or not isinstance(
                    widget := self.notification_widgets[notification.id][1],
                    SpinnerWidget,
                ):
                    [_, widget] = self.notification_widgets[notification.id] = (
                        notification,
                        SpinnerWidget(
                            font_size=dp(16),
                            size_hint=(None, None),
                            pos_hint={'center_y': 0.5},
                            height=dp(16),
                            width=dp(16),
                        ),
                    )
                widget.color = notification.color
            else:
                if notification.id not in self.notification_widgets or not isinstance(
                    widget := self.notification_widgets[notification.id][1],
                    ProgressRingWidget,
                ):
                    [_, widget] = self.notification_widgets[notification.id] = (
                        notification,
                        ProgressRingWidget(
                            background_color=(0.3, 0.3, 0.3, 1),
                            height=dp(16),
                            band_width=dp(7),
                            size_hint=(None, None),
                            pos_hint={'center_y': 0.5},
                        ),
                    )
                widget.progress = notification.progress
                widget.color = notification.color

        for id in set(self.notification_widgets) - {
            notification.id for notification in notifications
        }:
            del self.notification_widgets[id]

        self.progress_layout.clear_widgets()

        for item in sorted(
            self.notification_widgets.values(),
            key=lambda item: item[0].timestamp,
        ):
            self.progress_layout.add_widget(item[1])

        self.progress_layout.width = self.progress_layout.minimum_width

    @mainthread
    def handle_is_header_visible_change(
        self: MenuAppHeader,
        is_header_visible: bool,  # noqa: FBT001
    ) -> None:
        if is_header_visible:
            if self.header_content not in self.header_layout.children:
                self.header_layout.add_widget(self.header_content)
        elif self.header_content in self.header_layout.children:
            self.header_layout.remove_widget(self.header_content)

    @mainthread
    def handle_is_recording_actions_change(
        self: MenuAppHeader,
        is_recording: bool,  # noqa: FBT001
    ) -> None:
        if is_recording:
            if self.recording_sign not in self.header_content.children:
                self.recording_sign.color = (0, 0, 1, 1)
                self.header_content.add_widget(self.recording_sign)
                self.sign_animation.start(self.recording_sign)
        elif self.recording_sign in self.header_content.children:
            self.header_content.remove_widget(self.recording_sign)
            self.sign_animation.cancel(self.recording_sign)

    @mainthread
    def handle_is_replaying_change(
        self: MenuAppHeader,
        is_replaying: bool,  # noqa: FBT001
    ) -> None:
        if is_replaying:
            if self.replaying_sign not in self.header_content.children:
                self.header_content.add_widget(self.replaying_sign)
                self.sign_animation.start(self.replaying_sign)
        elif self.replaying_sign in self.header_content.children:
            self.header_content.remove_widget(self.replaying_sign)
            self.sign_animation.cancel(self.replaying_sign)

    @mainthread
    def handle_is_recording_audio_change(
        self: MenuAppHeader,
        is_recording: bool,  # noqa: FBT001
    ) -> None:
        if is_recording:
            if self.recording_sign not in self.header_content.children:
                self.recording_sign.color = (0, 1, 0, 1)
                self.header_content.add_widget(self.recording_sign)
                self.sign_animation.start(self.recording_sign)
        elif self.recording_sign in self.header_content.children:
            self.header_content.remove_widget(self.recording_sign)
            self.sign_animation.cancel(self.recording_sign)

    @cached_property
    def header(self: MenuAppHeader) -> Widget | None:
        self.header_content = RelativeLayout()

        original_header = super().header
        if isinstance(original_header, Label):
            original_header.bind(size=original_header.setter('text_size'))
            original_header.halign = 'center'
            original_header.valign = 'center'
            original_header.shorten = True

        if not original_header:
            return None
        original_header.pos = (0, 0)
        self.header_content.add_widget(original_header)

        self.progress_layout = BoxLayout(
            orientation='horizontal',
            padding=dp(4),
            spacing=dp(2),
        )
        self.header_content.add_widget(self.progress_layout)

        self.recording_sign = Label(
            text='󰑊',
            font_size=dp(20),
            color=(1, 0, 0, 1),
            pos_hint={'right': 1},
            size_hint=(None, 1),
        )
        self.recording_sign.bind(texture_size=self.recording_sign.setter('size'))
        self.replaying_sign = Label(
            text='󰑙',
            font_size=dp(20),
            color=(0, 1, 0, 1),
            pos_hint={'right': 1},
            size_hint=(None, 1),
        )
        self.replaying_sign.bind(texture_size=self.replaying_sign.setter('size'))
        self.sign_animation = (
            Animation(opacity=1, duration=0.1)
            + Animation(duration=1)
            + Animation(opacity=0, duration=0.1)
            + Animation(duration=0.5)
        )
        self.sign_animation.repeat = True

        self.notification_widgets = {}

        self.header_layout = BoxLayout()
        self.header_layout.add_widget(self.header_content)

        store.autorun(
            lambda state: [
                notification
                for notification in state.notifications.notifications
                if notification.progress is not None
            ],
            options=AutorunOptions(keep_ref=False),
        )(self.set_notification_widgets)

        store.autorun(
            lambda state: state.main.is_header_visible,
            options=AutorunOptions(keep_ref=False),
        )(self.handle_is_header_visible_change)

        store.autorun(
            lambda state: state.main.is_recording,
            options=AutorunOptions(keep_ref=False),
        )(self.handle_is_recording_actions_change)

        store.autorun(
            lambda state: state.main.is_replaying,
            options=AutorunOptions(keep_ref=False),
        )(self.handle_is_replaying_change)

        store.autorun(
            lambda state: state.audio.is_recording,
            options=AutorunOptions(keep_ref=False),
        )(self.handle_is_recording_audio_change)

        return self.header_layout



"""Setup the service."""

from __future__ import annotations

import asyncio
import errno
from datetime import UTC, datetime
from typing import TYPE_CHECKING, TypeVar

import adafruit_pct2075
import adafruit_veml7700
import board
from tenacity import retry, retry_if_exception, stop_after_attempt, wait_fixed

from ubo_app.logger import logger
from ubo_app.store.main import store
from ubo_app.store.services.sensors import Sensor, SensorsReportReadingAction
from ubo_app.utils.async_ import create_task
from ubo_app.utils.eeprom import get_eeprom_data
from ubo_app.utils.error_handlers import report_service_error

if TYPE_CHECKING:
    from adafruit_rgb_display.rgb import busio

    from ubo_app.utils.types import Subscriptions

temperature_sensor: adafruit_pct2075.PCT2075 | None = None
light_sensor: adafruit_veml7700.VEML7700 | None = None


def read_sensors(_: float | None = None) -> None:
    """Read the sensors."""
    temperature = 0.0 if temperature_sensor is None else temperature_sensor.temperature
    light = 0.0 if light_sensor is None else light_sensor.lux
    store.dispatch(
        SensorsReportReadingAction(
            sensor=Sensor.TEMPERATURE,
            reading=temperature,
            timestamp=datetime.now(tz=UTC),
        ),
        SensorsReportReadingAction(
            sensor=Sensor.LIGHT,
            reading=light,
            timestamp=datetime.now(tz=UTC),
        ),
    )


T = TypeVar('T', bound=adafruit_pct2075.PCT2075 | adafruit_veml7700.VEML7700)


@retry(
    stop=stop_after_attempt(5),
    wait=wait_fixed(1),
    retry=retry_if_exception(
        lambda e: isinstance(e, OSError) and e.errno == errno.EIO,
    ),
)
def _initialize_device(cls: type[T], address: int, i2c: busio.I2C) -> T:
    """Initialize the I2C device."""
    return cls(i2c, address)


async def _monitor_sensors(end_event: asyncio.Event) -> None:
    while not end_event.is_set():
        read_sensors()
        await asyncio.sleep(1)


def init_service() -> Subscriptions:
    """Initialize the service."""
    eeprom_data = get_eeprom_data()

    global temperature_sensor, light_sensor  # noqa: PLW0603

    i2c = board.I2C()
    try:
        if (
            (temp := eeprom_data.get('temperature')) is not None
            and temp.get('model', '').upper() == 'PCT2075'
            and temp.get('bus_address')
        ):
            temperature_sensor = _initialize_device(
                adafruit_pct2075.PCT2075,
                int(temp['bus_address'], 16),
                i2c,
            )
    except Exception:
        logger.exception('Error initializing temperature sensor')
        report_service_error()

    try:
        if (
            (ambient := eeprom_data.get('ambient')) is not None
            and ambient.get('model', '').upper() == 'VEML7700'
            and ambient.get('bus_address')
        ):
            light_sensor = _initialize_device(
                adafruit_veml7700.VEML7700,
                int(ambient['bus_address'], 16),
                i2c,
            )
            light_sensor.light_integration_time = adafruit_veml7700.VEML7700.ALS_50MS
    except Exception:
        logger.exception('Error initializing light sensor')
        report_service_error()

    read_sensors()

    end_event = asyncio.Event()
    create_task(_monitor_sensors(end_event))

    return [end_event.set]



"""Reset button module."""

import pwd
import subprocess
import time

from ubo_app.constants import USERNAME
from ubo_app.logger import get_logger
from ubo_app.system.system_manager.led import LEDManager

logger = get_logger('system-manager')

RESET_BUTTON_PIN = 27
LONG_PRESS_TIME = 5
SHORT_PRESS_TIME = 1
RAPID_PRESS_TIME = 1
RAPID_PRESS_COUNT = 3


def setup_reset_button(led_manager: LEDManager) -> None:
    """Set the reset button."""
    from gpiozero import Button

    reset_button = Button(RESET_BUTTON_PIN)
    state = {}
    state['latest_releases'] = []

    def reset_button_pressed(_: int) -> None:
        state['press_time'] = time.time()

    def reset_button_released(_: int) -> None:
        release_time = time.time()
        state['latest_releases'] = [
            release_time,
        ] + [t for t in state['latest_releases'] if release_time - t < RAPID_PRESS_TIME]
        if 'press_time' in state:
            duration = release_time - state['press_time']
            logger.info(
                'Reset button released',
                extra={
                    'duration': duration,
                    'press_time': state['press_time'],
                    'release_time': release_time,
                    'latest_releases': state['latest_releases'],
                    'SHORT_PRESS_TIME': SHORT_PRESS_TIME,
                    'LONG_PRESS_TIME': LONG_PRESS_TIME,
                    'RAPID_PRESS_TIME': RAPID_PRESS_TIME,
                    'RAPID_PRESS_COUNT': RAPID_PRESS_COUNT,
                },
            )

            if duration > LONG_PRESS_TIME:
                # Restart the pod
                led_manager.run_command_thread_safe(
                    ['blink', '255', '0', '0', '200', '2'],
                )
                subprocess.run(
                    ['/usr/bin/env', 'systemctl', 'reboot', '-i'],
                    check=False,
                )
            elif duration > SHORT_PRESS_TIME:
                # Restart the ubo service
                led_manager.run_command_thread_safe(
                    ['blink', '255', '255', '0', '200', '1'],
                )
                uid = pwd.getpwnam(USERNAME).pw_uid
                subprocess.run(  # noqa: S603
                    [
                        '/usr/bin/env',
                        'sudo',
                        f'XDG_RUNTIME_DIR=/run/user/{uid}',
                        '-u',
                        USERNAME,
                        'systemctl',
                        '--user',
                        'restart',
                        'ubo-app',
                    ],
                    check=False,
                )
                time.sleep(1)
                led_manager.run_initialization_loop()
            elif len(state['latest_releases']) == RAPID_PRESS_COUNT:
                # Kill the UBO process
                led_manager.run_command_thread_safe(
                    ['blink', '255', '0', '0', '200', '1'],
                )
                subprocess.run(
                    ['/usr/bin/env', 'killall', '-9', 'ubo'],
                    check=False,
                )
                time.sleep(1)
                led_manager.run_initialization_loop()

    reset_button.when_pressed = reset_button_pressed
    reset_button.when_released = reset_button_released



"""Camera settings menu pages."""

from __future__ import annotations

from ubo_gui.menu.types import ActionItem, HeadedMenu, SubMenuItem

from ubo_app.store.main import store
from ubo_app.store.services.camera import (
    CameraDetectAction,
    CameraSetIndexAction,
    CameraState,
)


@store.autorun(lambda state: state.camera)
def camera_settings_menu(state: CameraState) -> HeadedMenu:
    """Create camera selection menu with detected cameras.

    Args:
        state: Camera state from the store

    Returns:
        HeadedMenu with camera selection options

    """
    items = []

    # Add menu items for each detected camera
    for index in state.available_cameras:
        is_selected = index == state.selected_camera_index
        items.append(
            ActionItem(
                label=f'Camera {index}',
                icon='' if is_selected else '',  # noqa: RUF034
                background_color='#00ff00' if is_selected else None,
                color='#000000' if is_selected else None,
                action=lambda i=index: store.dispatch(CameraSetIndexAction(index=i)),
            ),
        )

    # Add detect cameras button
    items.append(
        ActionItem(
            label='Detect Cameras',
            icon='󰄄',
            action=lambda: store.dispatch(CameraDetectAction()),
        ),
    )

    return HeadedMenu(
        title='Camera Settings',
        heading='Select Camera Device',
        sub_heading=f'Current: Camera {state.selected_camera_index}'
        if state.available_cameras
        else 'No cameras detected',
        items=items,
        placeholder='No cameras detected. Click "Detect Cameras" to scan.'
        if not state.available_cameras
        else None,
    )


# Create the menu item for registration
CameraSettingsMenu = SubMenuItem(
    label='Camera',
    icon='',
    sub_menu=camera_settings_menu,
)



# ruff: noqa: D100, D101, D102, D107
from __future__ import annotations

import weakref
from functools import cached_property
from typing import TYPE_CHECKING

from debouncer import DebounceOptions, debounce
from ubo_gui.app import UboApp
from ubo_gui.menu.menu_widget import MenuPageWidget, MenuWidget
from ubo_gui.menu.stack_item import StackApplicationItem, StackItem, StackMenuItem
from ubo_gui.page import PageWidget
from ubo_gui.utils import mainthread_if_needed

from ubo_app.constants import DEBUG_MENU
from ubo_app.logger import logger
from ubo_app.menu_app.home_page import HomePage
from ubo_app.menu_app.menu_notification_handler import MenuNotificationHandler
from ubo_app.store.core.types import (
    CloseApplicationEvent,
    MenuChooseByIconEvent,
    MenuChooseByIndexEvent,
    MenuChooseByLabelEvent,
    MenuGoBackEvent,
    MenuGoHomeEvent,
    MenuScrollDirection,
    MenuScrollEvent,
    OpenApplicationEvent,
    SetAreEnclosuresVisibleAction,
    SetMenuPathAction,
)
from ubo_app.store.main import store
from ubo_app.store.services.notifications import NotificationsDisplayEvent
from ubo_app.store.ubo_actions import get_registered_application
from ubo_app.utils.gui import UboPageWidget

if TYPE_CHECKING:
    from collections.abc import Sequence

    from kivy.uix.widget import Widget
    from ubo_gui.menu.types import Item, Menu


class MenuWidgetWithHomePage(MenuWidget):
    @cached_property
    def home_page(self: MenuWidgetWithHomePage) -> HomePage:
        return HomePage(
            name='Page 1 0',
            padding_bottom=self.padding_bottom,
            padding_top=self.padding_top,
        )

    def _render_menu(self: MenuWidgetWithHomePage, menu: Menu) -> MenuPageWidget:
        if self.depth <= 1:
            self.home_page.items = self.current_menu_items
            self.current_screen = self.home_page
            return self.home_page
        return super()._render_menu(menu)


class MenuAppCentral(MenuNotificationHandler, UboApp):
    def __init__(self: MenuAppCentral, **kwargs: object) -> None:
        super().__init__(**kwargs)
        self.menu_widget = MenuWidgetWithHomePage(render_surroundings=True)

        self.menu_widget.bind(page_index=self.handle_page_index_change)
        self.menu_widget.bind(current_menu=self.handle_page_index_change)
        self.menu_widget.bind(title=self.handle_title_change)
        self.menu_widget.bind(stack=self.handle_stack_change)

        if DEBUG_MENU:
            menu_representation = 'Menu:\n' + repr(self.menu_widget)
            self.menu_widget.bind(stack=lambda *_: logger.info(menu_representation))

        _self = weakref.ref(self)

        @store.autorun(lambda state: state.main.menu)
        @debounce(0.1, DebounceOptions(leading=True, trailing=True, time_window=0.1))
        def _(menu: Menu | None) -> None:
            self = _self()
            if not self or not menu:
                return
            self.menu_widget.set_root_menu(menu)

    def build(self) -> Widget | None:
        root = super().build()
        if root:
            self.menu_widget.padding_top = root.ids.header_layout.height
            self.menu_widget.padding_bottom = root.ids.footer_layout.height

        return root

    def handle_page_index_change(
        self: MenuAppCentral,
        *_: object,
    ) -> None:
        store.dispatch(
            SetAreEnclosuresVisibleAction(
                is_header_visible=self.menu_widget.page_index == 0,
                is_footer_visible=self.menu_widget.page_index
                >= self.menu_widget.pages - 1,
            ),
        )

    def handle_title_change(self: MenuAppCentral, _: MenuWidget, title: str) -> None:
        self.root.title = title

    def handle_stack_change(
        self: MenuAppCentral,
        _: MenuWidget,
        stack: list[StackItem],
    ) -> None:
        store.dispatch(
            SetMenuPathAction(
                path=[
                    stack_item.selection.key
                    for stack_item in stack
                    if isinstance(stack_item, StackMenuItem) and stack_item.selection
                ],
                depth=len(stack),
            ),
        )

    @cached_property
    def central(self: MenuAppCentral) -> Widget | None:
        """Build the main menu and initiate it."""
        self.root.title = self.menu_widget.title

        store.subscribe_event(
            NotificationsDisplayEvent,
            self.display_notification,
            keep_ref=False,
        )

        store.subscribe_event(
            OpenApplicationEvent,
            self.open_application,
            keep_ref=False,
        )
        store.subscribe_event(
            CloseApplicationEvent,
            self.close_application,
            keep_ref=False,
        )
        store.subscribe_event(
            MenuGoHomeEvent,
            self.go_home,
            keep_ref=False,
        )
        store.subscribe_event(
            MenuGoBackEvent,
            self.go_back,
            keep_ref=False,
        )
        store.subscribe_event(
            MenuChooseByIconEvent,
            self.select_by_icon,
            keep_ref=False,
        )
        store.subscribe_event(
            MenuChooseByLabelEvent,
            self.select_by_label,
            keep_ref=False,
        )
        store.subscribe_event(
            MenuChooseByIndexEvent,
            self.select_by_index,
            keep_ref=False,
        )
        store.subscribe_event(
            MenuScrollEvent,
            self.scroll,
            keep_ref=False,
        )

        return self.menu_widget

    @mainthread_if_needed
    def open_application(self: MenuAppCentral, event: OpenApplicationEvent) -> None:
        application = get_registered_application(event.application_id)
        self.menu_widget.open_application(
            application(*event.initialization_args, **event.initialization_kwargs),
        )

    def close_application(self: MenuAppCentral, event: CloseApplicationEvent) -> None:
        for item in self.menu_widget.stack:
            if (
                event.application_instance_id
                and isinstance(item, StackApplicationItem)
                and hasattr(item.application, 'id')
                and isinstance(item.application, UboPageWidget)
                and item.application.id == event.application_instance_id
            ):
                self.menu_widget.close_application(item.application)

    def go_home(self: MenuAppCentral, _: MenuGoHomeEvent) -> None:
        self.menu_widget.go_home()

    def go_back(self: MenuAppCentral, _: MenuGoBackEvent) -> None:
        self.menu_widget.go_back()

    def select_by_icon(self: MenuAppCentral, event: MenuChooseByIconEvent) -> None:
        current_page = self.menu_widget.current_screen
        if current_page is None:
            msg = 'No current page'
            raise ValueError(msg)
        if not isinstance(current_page, PageWidget):
            msg = 'Current page is not a StackMenuItem'
            raise TypeError(msg)
        if current_page.items is None:
            msg = 'No items in current page'
            raise TypeError(msg)
        items: Sequence[Item | None] = current_page.items
        filtered_items = [item for item in items if item and item.icon == event.icon]
        if not filtered_items:
            msg = f'No item with icon "{event.icon}"'
            raise ValueError(msg)
        if len(filtered_items) > 1:
            msg = (
                f'Expected 1 item with icon "{event.icon}", found '
                f'"{len(filtered_items)}"'
            )
            raise ValueError(msg)
        self.menu_widget.select_item(filtered_items[0], parent=self.menu_widget.top)

    def select_by_label(
        self: MenuAppCentral,
        event: MenuChooseByLabelEvent,
    ) -> None:
        from ubo_gui.page import PageWidget

        current_page = self.menu_widget.current_screen
        if current_page is None:
            msg = 'No current page'
            raise ValueError(msg)
        if not isinstance(current_page, PageWidget):
            msg = 'Current page is not a StackMenuItem'
            raise TypeError(msg)
        if current_page.items is None:
            msg = 'No items in current page'
            raise TypeError(msg)
        items: Sequence[Item | None] = current_page.items
        filtered_items = [item for item in items if item and item.label == event.label]
        if not filtered_items:
            msg = f'No item with label "{event.label}"'
            raise ValueError(msg)
        if len(filtered_items) > 1:
            msg = (
                f'Expected 1 item with label "{event.label}", found '
                f'"{len(filtered_items)}"'
            )
            raise ValueError(msg)
        self.menu_widget.select_item(filtered_items[0], parent=self.menu_widget.top)

    def select_by_index(
        self: MenuAppCentral,
        event: MenuChooseByIndexEvent,
    ) -> None:
        self.menu_widget.select(event.index)

    def scroll(self: MenuAppCentral, event: MenuScrollEvent) -> None:
        if event.direction == MenuScrollDirection.UP:
            self.menu_widget.go_up()
        elif event.direction == MenuScrollDirection.DOWN:
            self.menu_widget.go_down()



# ruff: noqa: D100, D101, D103
from __future__ import annotations

import asyncio
import json
import socket
import subprocess
from typing import Literal, TypedDict

from constants_ import CODE_BINARY_PATH
from debouncer import DebounceOptions, debounce

from ubo_app.colors import DANGER_COLOR
from ubo_app.logger import logger
from ubo_app.store.main import store
from ubo_app.store.services.notifications import (
    Chime,
    Notification,
    NotificationDisplayType,
    NotificationsAddAction,
)
from ubo_app.store.services.vscode import (
    VSCodeSetPendingAction,
    VSCodeSetStatusAction,
    VSCodeStatus,
)


class TunnelStatus(TypedDict):
    tunnel: Literal['Connected', 'Disconnected']
    name: str | None


class TunnelServiceStatus(TypedDict):
    service_installed: bool
    tunnel: TunnelStatus


@debounce(
    wait=1,
    options=DebounceOptions(leading=True, trailing=False, time_window=1),
)
async def check_status() -> None:
    is_binary_installed = CODE_BINARY_PATH.exists()
    status_data: TunnelServiceStatus | None = None
    is_logged_in = False
    try:
        if is_binary_installed:
            process = await asyncio.create_subprocess_exec(
                CODE_BINARY_PATH.as_posix(),
                'tunnel',
                '--accept-server-license-terms',
                'status',
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.DEVNULL,
            )
            await asyncio.wait_for(process.wait(), timeout=3)
            if process.returncode is None:
                process.kill()
            if process.stdout and process.returncode == 0:
                output = await process.stdout.read()
                status_data = json.loads(output)
    except (subprocess.CalledProcessError, TimeoutError):
        store.dispatch(
            NotificationsAddAction(
                notification=Notification(
                    id='vscode:error:status',
                    title='VSCode',
                    content='Failed to get status: "status" subcommand',
                    display_type=NotificationDisplayType.STICKY,
                    color=DANGER_COLOR,
                    icon='󰜺',
                    chime=Chime.FAILURE,
                ),
            ),
        )
        raise
    else:
        try:
            if is_binary_installed:
                process = await asyncio.create_subprocess_exec(
                    CODE_BINARY_PATH.as_posix(),
                    'tunnel',
                    '--accept-server-license-terms',
                    'user',
                    'show',
                    stdout=asyncio.subprocess.DEVNULL,
                    stderr=asyncio.subprocess.DEVNULL,
                )
                await asyncio.wait_for(process.wait(), timeout=3)
                if process.returncode is None:
                    process.kill()
                is_logged_in = process.returncode == 0
        except (subprocess.CalledProcessError, TimeoutError):
            store.dispatch(
                NotificationsAddAction(
                    notification=Notification(
                        id='vscode:error:user',
                        title='VSCode',
                        content='Failed to get status: "user show" subcommand',
                        display_type=NotificationDisplayType.STICKY,
                        color=DANGER_COLOR,
                        icon='󰜺',
                        chime=Chime.FAILURE,
                    ),
                ),
            )
            raise
    logger.debug(
        'Checked VSCode Tunnel Status',
        extra={
            'status': status_data,
            'is_logged_in': is_logged_in,
            'is_binary_installed': is_binary_installed,
        },
    )
    store.dispatch(
        VSCodeSetStatusAction(
            is_binary_installed=is_binary_installed,
            is_logged_in=is_logged_in,
            status=None
            if status_data is None
            else VSCodeStatus(
                is_service_installed=status_data['service_installed'],
                is_running=status_data['tunnel'] is not None
                and status_data['tunnel']['tunnel'] == 'Connected',
                name=None
                if status_data['tunnel'] is None
                else status_data['tunnel']['name'],
            ),
        ),
    )


async def set_name() -> None:
    store.dispatch(VSCodeSetPendingAction())
    try:
        hostname = socket.gethostname()
        process = await asyncio.create_subprocess_exec(
            CODE_BINARY_PATH,
            'tunnel',
            '--accept-server-license-terms',
            'rename',
            hostname,
            stdout=asyncio.subprocess.DEVNULL,
            stderr=asyncio.subprocess.DEVNULL,
        )
        await asyncio.wait_for(process.wait(), timeout=3)
        if process.returncode is None:
            process.kill()
    except (subprocess.CalledProcessError, TimeoutError):
        store.dispatch(
            NotificationsAddAction(
                notification=Notification(
                    id='vscode:error:rename',
                    title='VSCode',
                    content='Failed to setup: renaming the tunnel',
                    display_type=NotificationDisplayType.STICKY,
                    color=DANGER_COLOR,
                    icon='󰜺',
                    chime=Chime.FAILURE,
                ),
            ),
        )
        raise
    finally:
        await check_status()


async def install_service() -> None:
    store.dispatch(VSCodeSetPendingAction())
    try:
        process = await asyncio.create_subprocess_exec(
            CODE_BINARY_PATH,
            'tunnel',
            '--accept-server-license-terms',
            'service',
            'install',
        )
        await asyncio.wait_for(process.wait(), timeout=3)
        if process.returncode is None:
            process.kill()
    except (subprocess.CalledProcessError, TimeoutError):
        store.dispatch(
            NotificationsAddAction(
                notification=Notification(
                    id='vscode:error:install',
                    title='VSCode',
                    content='Failed to setup: installing service',
                    display_type=NotificationDisplayType.STICKY,
                    color=DANGER_COLOR,
                    icon='󰜺',
                    chime=Chime.FAILURE,
                ),
            ),
        )
        raise
    finally:
        await check_status()


async def uninstall_service() -> None:
    store.dispatch(VSCodeSetPendingAction())
    try:
        process = await asyncio.create_subprocess_exec(
            CODE_BINARY_PATH,
            'tunnel',
            '--accept-server-license-terms',
            'service',
            'uninstall',
            stdout=asyncio.subprocess.DEVNULL,
            stderr=asyncio.subprocess.DEVNULL,
        )
        await asyncio.wait_for(process.wait(), timeout=3)
        if process.returncode is None:
            process.kill()
    except (subprocess.CalledProcessError, TimeoutError):
        store.dispatch(
            NotificationsAddAction(
                notification=Notification(
                    id='vscode:error:uninstall',
                    title='VSCode',
                    content='Failed to setup: uninstalling service',
                    display_type=NotificationDisplayType.STICKY,
                    color=DANGER_COLOR,
                    icon='󰜺',
                    chime=Chime.FAILURE,
                ),
            ),
        )
        raise
    finally:
        await check_status()


async def restart() -> None:
    store.dispatch(VSCodeSetPendingAction())
    try:
        process = await asyncio.create_subprocess_exec(
            CODE_BINARY_PATH,
            'tunnel',
            '--accept-server-license-terms',
            'restart',
            stdout=asyncio.subprocess.DEVNULL,
            stderr=asyncio.subprocess.DEVNULL,
        )
        await asyncio.wait_for(process.wait(), timeout=3)
        if process.returncode is None:
            process.kill()
    except (subprocess.CalledProcessError, TimeoutError):
        store.dispatch(
            NotificationsAddAction(
                notification=Notification(
                    id='vscode:error:restart',
                    title='VSCode',
                    content='Failed to restart process',
                    display_type=NotificationDisplayType.STICKY,
                    color=DANGER_COLOR,
                    icon='󰜺',
                    chime=Chime.FAILURE,
                ),
            ),
        )
        raise
    finally:
        await check_status()



"""Docker image management."""

from __future__ import annotations

from dataclasses import replace

import docker
import docker.errors
from calculate_progress import LayerProgressTracker
from docker_images import IMAGES

from ubo_app.colors import DANGER_COLOR
from ubo_app.constants import DOCKER_CREDENTIALS_TEMPLATE_SECRET_ID
from ubo_app.logger import logger
from ubo_app.store.main import store
from ubo_app.store.services.docker import (
    DockerImageFetchEvent,
    DockerImageRemoveEvent,
    DockerImageSetStatusAction,
    DockerItemStatus,
)
from ubo_app.store.services.notifications import (
    Chime,
    Notification,
    NotificationDisplayType,
    NotificationsAddAction,
)
from ubo_app.utils import secrets
from ubo_app.utils.async_ import to_thread

# Notification IDs for progress tracking
DOCKER_FETCH_PROGRESS_NOTIFICATION_ID = 'docker:fetch_progress:{}'


def _create_fetch_progress_notification(
    image_id: str,
    content: str = 'Starting download...',
    progress: float | None = 0.0,
) -> Notification:
    """Create a fetch progress notification for an image."""
    return Notification(
        id=DOCKER_FETCH_PROGRESS_NOTIFICATION_ID.format(image_id),
        title=f'{IMAGES[image_id].label}',
        content=content,
        display_type=NotificationDisplayType.BACKGROUND,
        icon=IMAGES[image_id].icon,
        show_dismiss_action=False,
        progress=progress,
    )


@store.with_state(lambda state: state.docker.service.usernames)
def fetch_image(  # noqa: C901
    usernames: dict[str, str],
    event: DockerImageFetchEvent,
) -> None:
    """Fetch a container image."""
    id = event.image

    def act() -> None:
        # Create base notification
        base_notification = _create_fetch_progress_notification(id)

        # Dispatch initial status and notification
        store.dispatch(
            DockerImageSetStatusAction(
                image=id,
                status=DockerItemStatus.FETCHING,
            ),
            NotificationsAddAction(notification=base_notification),
        )

        try:
            # Construct full image path with registry if specified
            image_entry = IMAGES[id]
            if image_entry.registry:
                full_image_path = f'{image_entry.registry}/{image_entry.path}'
            else:
                full_image_path = image_entry.path

            logger.info('Fetching image', extra={'full_path': full_image_path})
            docker_client = docker.from_env()

            # Login to registry if credentials are available
            for registry, username in usernames.items():
                if image_entry.registry == registry:
                    docker_client.login(
                        username=username,
                        password=secrets.read_secret(
                            DOCKER_CREDENTIALS_TEMPLATE_SECRET_ID.format(registry),
                        ),
                        registry=registry,
                    )

            # Pull the image with progress tracking
            logger.info(
                'Starting image pull',
                extra={'full_path': full_image_path, 'image': id},
            )
            pull_result = docker_client.api.pull(
                full_image_path,
                stream=True,
                decode=True,
            )

            # Initialize progress tracker
            progress_tracker = LayerProgressTracker(id)

            for line in pull_result:
                if 'status' in line:
                    layer_id = line.get('id', 'unknown')
                    status = line['status']
                    progress = line.get('progress', '')

                    # Update progress tracker
                    progress_tracker.update_layer(layer_id, status, progress)

                    # Update UI periodically (debounced)
                    if progress_tracker.should_update_ui():
                        current_progress = progress_tracker.calculate_progress()
                        store.dispatch(
                            NotificationsAddAction(
                                notification=replace(
                                    base_notification,
                                    content='Downloading image...',
                                    progress=current_progress,
                                ),
                            ),
                        )

                    # Log significant status changes
                    if status in (
                        'Downloading',
                        'Extracting',
                        'Pull complete',
                        'Already exists',
                    ):
                        logger.debug(
                            'Image pull progress',
                            extra={
                                'image': id,
                                'layer': layer_id,
                                'status': status,
                                'progress': progress,
                            },
                        )

                if 'error' in line:
                    error_msg = line['error']
                    logger.error(
                        'Image pull error',
                        extra={'image': id, 'error': error_msg},
                    )
                    msg = f'Pull failed: {error_msg}'
                    raise docker.errors.ImageNotFound(msg)

            logger.info('Image pull completed', extra={'image': id})
            docker_client.close()

            # Dispatch success notification
            store.dispatch(
                NotificationsAddAction(
                    notification=replace(
                        base_notification,
                        content='Download complete!',
                        progress=1.0,
                        show_dismiss_action=True,
                        display_type=NotificationDisplayType.FLASH,
                        dismiss_on_close=True,
                        chime=Chime.DONE,
                    ),
                ),
            )
        except docker.errors.DockerException as e:
            logger.exception(
                'Failed to fetch image',
                extra={'image': id, 'error': str(e)},
            )
            store.dispatch(
                DockerImageSetStatusAction(
                    image=id,
                    status=DockerItemStatus.ERROR,
                ),
                NotificationsAddAction(
                    notification=replace(
                        base_notification,
                        display_type=NotificationDisplayType.FLASH,
                        content='Download failed',
                        color=DANGER_COLOR,
                        show_dismiss_action=True,
                        progress=None,
                        chime=Chime.FAILURE,
                    ),
                ),
                NotificationsAddAction(
                    notification=Notification(
                        id=f'docker_fetch_error_{id}',
                        title=f'Failed to Fetch {IMAGES[id].label}',
                        content=f'Error: {str(e).split("(")[0].strip()}',
                        display_type=NotificationDisplayType.STICKY,
                        color=DANGER_COLOR,
                        icon=IMAGES[id].icon,
                        chime=Chime.FAILURE,
                    ),
                ),
            )
            raise

    to_thread(act)


def remove_image(event: DockerImageRemoveEvent) -> None:
    """Remove an image."""
    id = event.image
    logger.info(
        'remove_image called',
        extra={'image_id': id, 'path': IMAGES[id].full_path},
    )
    docker_client = docker.from_env()
    try:
        docker_client.images.remove(IMAGES[id].full_path, force=True)
        logger.info(
            'Image removed successfully - waiting for delete event',
            extra={'image_id': id},
        )
    except docker.errors.ImageNotFound:
        logger.warning(
            'Image already removed or not found',
            extra={'image_id': id, 'path': IMAGES[id].full_path},
        )
        # Still update status to NOT_AVAILABLE since image doesn't exist
        store.dispatch(
            DockerImageSetStatusAction(
                image=id,
                status=DockerItemStatus.NOT_AVAILABLE,
            ),
        )
    except docker.errors.DockerException as e:
        logger.exception(
            'Failed to remove image',
            extra={'image_id': id, 'error': str(e)},
        )
        raise
    finally:
        docker_client.close()



# ruff: noqa: D100, D101
from __future__ import annotations

import pathlib

from kivy.lang.builder import Builder
from kivy.metrics import dp
from kivy.properties import StringProperty

from ubo_app.utils.gui import UboPageWidget


class NotificationInfo(UboPageWidget):
    text: str = StringProperty()

    def go_down(self: NotificationInfo) -> None:
        """Scroll down the notification list."""
        self.ids.slider.animated_value -= dp(100)

    def go_up(self: NotificationInfo) -> None:
        """Scroll up the notification list."""
        self.ids.slider.animated_value += dp(100)


Builder.load_file(
    pathlib.Path(__file__).parent.joinpath('notification_info.kv').resolve().as_posix(),
)



"""Handle camera commands."""

from __future__ import annotations

import subprocess
from pathlib import Path

from ubo_app.logger import get_logger

logger = get_logger('system-manager')


def camera_handler(command: str, *args: str) -> str | None:
    """Handle camera commands."""
    if command == 'install_driver':
        make, model, variant = args[0], args[1], args[2]
        if make == 'arducam' and model == 'imx519':
            try:
                script = (
                    Path(__file__).parent.parent
                    / 'scripts/install_arducam_imx519.sh'
                ).read_bytes()
                process = subprocess.run(  # noqa: S603
                    ['/bin/bash', '-s', '--', variant],
                    input=script,
                    check=True,
                )
                process.check_returncode()
            except Exception:
                logger.exception('Error installing Arducam IMX519 driver')
                return 'error'
            else:
                return 'installed'
        logger.error(
            'Unknown camera driver: %s %s',
            make,
            model,
        )
        return 'error'
    if command == 'restore_default':
        try:
            script = (
                Path(__file__).parent.parent
                / 'scripts/restore_default_camera.sh'
            ).read_bytes()
            process = subprocess.run(
                ['/bin/bash', '-s'],
                input=script,
                check=True,
            )
            process.check_returncode()
        except Exception:
            logger.exception('Error restoring default camera configuration')
            return 'error'
        else:
            return 'restored'
    return None



"""Type definitions for the Redux store operations."""

from __future__ import annotations

import uuid
from dataclasses import field
from enum import StrEnum, auto
from typing import TYPE_CHECKING, TypeAlias

from immutable import Immutable
from redux import BaseAction, BaseEvent

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence

    from ubo_app.store.services.file_system import PathSelectorConfig
    from ubo_app.store.services.speech_synthesis import ReadableInformation


class InputMethod(StrEnum):
    """Input method."""

    CAMERA = auto()
    WEB_DASHBOARD = auto()
    PATH_SELECTOR = auto()


class InputResult(Immutable):
    """Input result."""

    data: Mapping[str, str]
    files: dict[str, bytes]
    method: InputMethod


class InputFieldType(StrEnum):
    """Enumeration of input field types."""

    LONG = 'long'
    TEXT = 'text'
    PASSWORD = 'password'  # noqa: S105
    NUMBER = 'number'
    CHECKBOX = 'checkbox'
    COLOR = 'color'
    SELECT = 'select'
    FILE = 'file'
    DATE = 'date'
    TIME = 'time'


class InputFieldDescription(Immutable):
    """Description of an input field in an input demand."""

    name: str
    label: str
    type: InputFieldType
    description: str | None = None
    title: str | None = None
    file_mimetype: str | None = None
    pattern: str | None = None
    default_value: str | None = None
    options: list[str] | None = None
    required: bool = False


class InputDescription(Immutable):
    """Description of an input demand."""

    input_method: InputMethod

    id: str = field(default_factory=lambda: uuid.uuid4().hex)
    title: str | None = None
    prompt: str | None = None


class WebUIInputDescription(InputDescription):
    """Description of a web UI input field."""

    input_method: InputMethod = InputMethod.WEB_DASHBOARD

    fields: Sequence[InputFieldDescription] | None = None


class QRCodeInputDescription(InputDescription):
    """Description of a QR code input field."""

    input_method: InputMethod = InputMethod.CAMERA

    instructions: ReadableInformation | None = None
    pattern: str | None = None


class PathInputDescription(InputDescription):
    """Description of a QR code input field."""

    input_method: InputMethod = InputMethod.PATH_SELECTOR

    selector_config: PathSelectorConfig


class InputAction(BaseAction):
    """Base class for input actions."""


AvailableInputDescription: TypeAlias = (
    WebUIInputDescription | QRCodeInputDescription | PathInputDescription
)


class InputDemandAction(InputAction):
    """Action for demanding input from the user."""

    description: AvailableInputDescription


class InputResolveAction(InputAction):
    """Base class for resolving input demands."""

    id: str


class InputCancelAction(InputResolveAction):
    """Action for cancelling an input demand."""


class InputProvideAction(InputResolveAction):
    """Action for reporting input from the user."""

    value: str
    result: InputResult | None


class InputResolveEvent(BaseEvent):
    """Base class for resolving input demands."""

    id: str


class InputCancelEvent(InputResolveEvent):
    """Event for cancelling an input demand."""


class InputProvideEvent(InputResolveEvent):
    """Event for reporting input from the user."""

    value: str
    result: InputResult | None



"""Sets up the display for the Raspberry Pi and manages its state."""

from __future__ import annotations

import asyncio
import datetime
from typing import TYPE_CHECKING

from ubo_gui.menu.types import HeadedMenu, SubMenuItem

from ubo_app.display import display
from ubo_app.logger import logger
from ubo_app.store.core.types import RegisterSettingAppAction, SettingsCategory
from ubo_app.store.main import store
from ubo_app.store.services.display import (
    DisplayBlankAction,
    DisplayBlankEvent,
    DisplayBlankTimeout,
    DisplaySetBlankTimeoutAction,
    DisplayState,
    DisplayUnblankEvent,
    DisplayUpdateActivityAction,
)
from ubo_app.store.ubo_actions import UboDispatchItem
from ubo_app.utils import IS_TEST_ENV
from ubo_app.utils.gui import (
    SELECTED_ITEM_PARAMETERS,
    UNSELECTED_ITEM_PARAMETERS,
)
from ubo_app.utils.persistent_store import register_persistent_store

if TYPE_CHECKING:
    from collections.abc import Sequence

    from ubo_gui.menu.types import Item

splash_screen = None

# Event to signal when the display is unblanked
# Python 3.10+ allows creation without running loop
_unblank_event = asyncio.Event()


async def monitor_inactivity() -> None:
    """Monitor user inactivity and blank screen after timeout."""
    logger.verbose('=== MONITOR INACTIVITY TASK STARTED ===')

    try:
        while True:
            # Get current display state using with_state
            @store.with_state(
                lambda state: (
                    state.display if hasattr(state, 'display') else None
                ),
            )
            def get_display_state(
                display_state: DisplayState | None,
            ) -> DisplayState | None:
                return display_state

            display_state = get_display_state()

            if display_state is None:
                logger.warning('Display state not available')
                await asyncio.sleep(10)
                continue

            # If display is already blanked, wait for unblank event instead of polling
            if display_state.is_blanked:
                logger.verbose(
                    'Display is blanked, waiting for unblank event',
                )
                _unblank_event.clear()
                await _unblank_event.wait()
                logger.verbose('Display unblanked, resuming inactivity monitoring')
                continue

            # Skip blanking if timeout is OFF
            timeout_seconds = display_state.selected_blank_timeout.get_timeout_seconds()
            if timeout_seconds is None:
                await asyncio.sleep(10)
                continue

            current_time = datetime.datetime.now(tz=datetime.UTC).timestamp()
            logger.verbose(
                'Inactivity check',
                extra={
                    'last_activity': display_state.last_activity_time,
                    'blank_timeout': timeout_seconds,
                    'is_blanked': display_state.is_blanked,
                    'current_time': current_time,
                },
            )

            if display_state.last_activity_time is not None:
                inactive_duration = current_time - display_state.last_activity_time
                logger.verbose(
                    'Checking inactivity duration',
                    extra={
                        'inactive_duration': inactive_duration,
                        'blank_timeout': timeout_seconds,
                        'will_blank': inactive_duration >= timeout_seconds,
                    },
                )
                if inactive_duration >= timeout_seconds:
                    logger.info(
                        'Screen blanking due to inactivity',
                        extra={'inactive_duration': inactive_duration},
                    )
                    store.dispatch(DisplayBlankAction())

            await asyncio.sleep(10)
            logger.verbose('=== INACTIVITY CHECK TICK ===')
    except Exception:
        logger.exception('Monitor inactivity task error')


def handle_blank_event(_: DisplayBlankEvent) -> None:
    """Handle screen blanking event."""
    logger.info('Blanking screen and turning off backlight')
    display.set_backlight(enabled=False)

def handle_unblank_event(_: DisplayUnblankEvent) -> None:
    """Handle screen unblanking event."""
    logger.info('Unblanking screen and turning on backlight')
    display.set_backlight(enabled=True)
    # Signal the inactivity monitor to wake up
    _unblank_event.set()


@store.autorun(lambda state: state.display.selected_blank_timeout)
def timeout_options(selected_timeout: DisplayBlankTimeout) -> Sequence[Item]:
    """Generate menu items for timeout selection."""
    return [
        UboDispatchItem(
            key=timeout.value,
            label=timeout.get_label(),
            store_action=DisplaySetBlankTimeoutAction(timeout=timeout),
            **(
                SELECTED_ITEM_PARAMETERS
                if selected_timeout == timeout
                else UNSELECTED_ITEM_PARAMETERS
            ),
        )
        for timeout in DisplayBlankTimeout
    ]


def init_service() -> None:
    """Initialize the display service."""
    from ubo_app.utils.async_ import create_task

    # Register persistent store
    register_persistent_store(
        'display:selected_blank_timeout',
        lambda state: state.display.selected_blank_timeout,
    )

    # Register settings menu
    store.dispatch(
        RegisterSettingAppAction(
            category=SettingsCategory.HARDWARE,
            priority=10,
            menu_item=SubMenuItem(
                label='Display',
                icon='󰍹',
                sub_menu=HeadedMenu(
                    title='Display Settings',
                    heading='Screen Timeout',
                    sub_heading='Select screen blank timeout',
                    items=timeout_options,
                ),
            ),
        ),
    )

    # Initialize activity tracking
    store.dispatch(DisplayUpdateActivityAction())
    logger.info('Display service initialized with activity tracking')

    store.subscribe_event(DisplayBlankEvent, handle_blank_event)
    store.subscribe_event(DisplayUnblankEvent, handle_unblank_event)

    # Don't start monitor task during tests to prevent screen blanking
    # during screenshots
    if not IS_TEST_ENV:
        create_task(monitor_inactivity())
        logger.info('Screen blanking monitor task started')
    else:
        logger.info('Screen blanking disabled during tests')




# ruff: noqa: D100, D103
from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING

from debouncer import DebounceOptions, debounce
from pages import create_wireless_connection, main
from wifi_manager import (
    get_connections,
    get_wifi_device,
    get_wifi_device_state,
    request_scan,
)

from ubo_app.colors import INFO_COLOR
from ubo_app.logger import logger
from ubo_app.store.core.types import (
    OpenApplicationAction,
    RegisterSettingAppAction,
    SettingsCategory,
)
from ubo_app.store.input.types import InputMethod
from ubo_app.store.main import store
from ubo_app.store.services.notifications import (
    Importance,
    Notification,
    NotificationDispatchItem,
    NotificationDisplayType,
    NotificationsAddAction,
)
from ubo_app.store.services.speech_synthesis import ReadableInformation
from ubo_app.store.services.wifi import (
    ConnectionState,
    WiFiInputConnectionAction,
    WiFiInputConnectionEvent,
    WiFiSetHasVisitedOnboardingAction,
    WiFiUpdateAction,
    WiFiUpdateRequestEvent,
)
from ubo_app.utils import IS_UBO_POD
from ubo_app.utils.async_ import create_task
from ubo_app.utils.network import get_saved_wifi_ssids, has_gateway
from ubo_app.utils.persistent_store import (
    read_from_persistent_store,
    register_persistent_store,
)

if TYPE_CHECKING:
    from ubo_app.utils.types import Subscriptions


@debounce(
    wait=0.5,
    options=DebounceOptions(leading=True, trailing=False, time_window=0.5),
)
async def update_wifi_list(_: WiFiUpdateRequestEvent | None = None) -> None:
    connections = await get_connections()

    store.dispatch(
        WiFiUpdateAction(
            connections=connections,
            state=await get_wifi_device_state(),
            current_connection=next(
                (
                    connection
                    for connection in connections
                    if connection.state is ConnectionState.CONNECTED
                ),
                None,
            ),
        ),
    )


async def setup_listeners() -> None:
    wifi_device = await get_wifi_device()
    if not wifi_device:
        return

    async for _ in wifi_device.properties_changed:
        create_task(update_wifi_list())


async def _check_connection() -> None:
    """Dispatch the Wi-Fi input action if needed."""
    await asyncio.sleep(2)
    logger.info(
        'Checking Wi-Fi',
        extra={
            'has_gateway': await has_gateway(),
            'saved_wifi_ssids': await get_saved_wifi_ssids(),
        },
    )
    onboarding_notification = Notification(
        title='No internet connection',
        content='Press middle button "󱚾" to add WiFi network',
        importance=Importance.MEDIUM,
        icon='󱚵',
        display_type=NotificationDisplayType.STICKY,
        actions=[
            NotificationDispatchItem(
                store_action=OpenApplicationAction(
                    application_id='wifi:create-connection-page',
                ),
                icon='󱚾',
                background_color=INFO_COLOR,
                dismiss_notification=True,
            ),
        ],
        extra_information=ReadableInformation(
            text='Press middle button to add a WiFi connection.\n'
            'If you dismiss this, you can always add WiFi through Settings → Network → '
            'WiFi.',
            piper_text='Press middle button to add a WiFi connection. '
            'If you dismiss this, you can always add WiFi through Settings menu, by '
            'navigating to Network, and then WiFi.',
            picovoice_text='Press middle button to add a {WiFi|W AY F AY} connection. '
            'If you dismiss this, you can always add {WiFi|W AY F AY} through Settings '
            '→ Network → {WiFi|W AY F AY}.',
        ),
        color=INFO_COLOR,
    )
    if not await has_gateway() and not await get_saved_wifi_ssids():
        if IS_UBO_POD:
            if not read_from_persistent_store(
                key='wifi_has_visited_onboarding',
                default=False,
            ):
                logger.info('No network connection found, showing WiFi onboarding.')
                store.dispatch(
                    NotificationsAddAction(
                        notification=onboarding_notification,
                    ),
                    WiFiSetHasVisitedOnboardingAction(has_visited_onboarding=True),
                )
        else:
            logger.info('No network connection found, prompting for Wi-Fi input.')
            store.dispatch(WiFiInputConnectionAction())


def init_service() -> Subscriptions:
    create_task(update_wifi_list())
    create_task(setup_listeners())

    register_persistent_store(
        'wifi_has_visited_onboarding',
        lambda state: state.wifi.has_visited_onboarding,
    )

    store.dispatch(
        RegisterSettingAppAction(
            priority=2,
            category=SettingsCategory.NETWORK,
            menu_item=main.WiFiMainMenu,
        ),
    )

    create_task(_check_connection())

    return [
        store.subscribe_event(WiFiUpdateRequestEvent, request_scan),
        store.subscribe_event(
            WiFiInputConnectionEvent,
            lambda: create_wireless_connection.input_wifi_connection(
                input_methods=(InputMethod.WEB_DASHBOARD,),
            ),
        ),
    ]



# ruff: noqa: D100, D101
from __future__ import annotations

from enum import StrEnum, auto
from typing import TYPE_CHECKING

from immutable import Immutable
from redux import BaseAction

if TYPE_CHECKING:
    from datetime import datetime


class SensorsAction(BaseAction): ...


class Sensor(StrEnum):
    TEMPERATURE = auto()
    LIGHT = auto()


class SensorsReportReadingAction(SensorsAction):
    sensor: Sensor
    reading: float
    timestamp: datetime


class SensorState(Immutable):
    value: float | None = None


class SensorsState(Immutable):
    temperature: SensorState = SensorState(value=None)
    light: SensorState = SensorState(value=None)



"""Compatibility layer for different environments."""

from __future__ import annotations

import asyncio
import signal
import subprocess
from pathlib import Path
from typing import TYPE_CHECKING, Any

import numpy as np
from fake import Fake
from redux import FinishAction

from ubo_app.utils import IS_TEST_ENV
from ubo_app.utils.eeprom import EMPTY_EEPROM_DATA

if TYPE_CHECKING:
    from ubo_gui.menu.types import Callable


original_subprocess_run = subprocess.run


def _fake_subprocess_run(
    command: list[str],
    *args: Any,  # noqa: ANN401
    **kwargs: Any,  # noqa: ANN401
) -> object:
    if any(i in command[0] for i in ('reboot', 'poweroff')):
        return Fake()
    return original_subprocess_run(command, *args, **kwargs)


original_asyncio_create_subprocess_exec = asyncio.create_subprocess_exec


async def _fake_create_subprocess_exec(
    *_args: str,
    **kwargs: Any,  # noqa: ANN401
) -> object:
    command = _args[0]
    args = _args[1:]

    if command == '/usr/bin/env':
        command, *args = args
    if isinstance(command, Path):
        command = command.as_posix()
    if any(i in command for i in ('reboot', 'poweroff', 'ir-ctl')):
        return Fake()
    if command in {'curl', 'tar'} or command.endswith('/code'):
        return await original_asyncio_create_subprocess_exec(*_args, **kwargs)
    if command == 'systemctl':
        if args[0] in 'is-enabled':
            return Fake(
                _Fake__return_value=Fake(_Fake__await_value=(b'enabled', b'')),
            )
        if args[0] == 'is-active':
            return Fake(
                _Fake__return_value=Fake(_Fake__await_value=(b'active', b'')),
            )
    if command == 'dpkg-query' and args[-1] in (
        'docker',
        'raspberrypi-ui-mods',
        'rpi-connect',
    ):
        return Fake(
            _Fake__return_value=Fake(_Fake__await_value=(b'install ok installed', b'')),
        )

    return await original_asyncio_create_subprocess_exec(*_args, **kwargs)


def setup() -> None:
    """Set up for different environments."""
    import sys
    import time

    from ubo_app.utils import IS_RPI

    if not IS_RPI:
        sys.modules['adafruit_irremote'] = Fake(
            _Fake__attrs={
                'GenericDecode': Fake(
                    _Fake__return_value=Fake(
                        _Fake__attrs={
                            'read_pulses': lambda *args, **kwargs: (args, kwargs)
                            and time.sleep(1),
                        },
                    ),
                ),
            },
        )
        sys.modules['adafruit_rgb_display.st7789'] = Fake()
        sys.modules['alsaaudio'] = Fake(_Fake__attrs={'cards': lambda: ['wm8960']})
        sys.modules['apt'] = Fake()
        sys.modules['board'] = Fake()
        sys.modules['digitalio'] = Fake()
        sys.modules['pulsectl'] = Fake()
        sys.modules['pulseio'] = Fake()
        sys.modules['sdbus'] = Fake()
        sys.modules['sdbus_async'] = Fake()
        sys.modules['sdbus_async.networkmanager'] = Fake()
        sys.modules['sdbus_async.networkmanager.enums'] = Fake()
        sys.modules['picamera2.picamera2'] = Fake(
            _Fake__attrs={
                'capture_array': Fake(
                    _Fake__return_value=np.zeros((1, 1, 3), dtype=np.uint8),
                ),
            },
        )
        sys.modules['sdbus.utils.inspect'] = Fake(
            _Fake__attrs={
                'inspect_dbus_path': lambda obj: obj._dbus.object_path,  # noqa: SLF001
            },
        )
        subprocess.run = _fake_subprocess_run

        asyncio.create_subprocess_exec = _fake_create_subprocess_exec

        asyncio.open_unix_connection = (
            Fake(_Fake__return_value=Fake(_Fake__await_value=(Fake(), Fake()))),
        )

        from ubo_app.utils import eeprom

        eeprom.get_eeprom_data = lambda: {
            **EMPTY_EEPROM_DATA,
            'speakers': {'model': 'wm8960'},
        }

        from ubo_app.utils import monitor_unit

        async def fake_monitor_unit(unit: str, callback: Callable[[str], None]) -> None:
            callback(
                'inactive' if unit in ('ubo-update.service') else 'active',
            )

        monitor_unit.monitor_unit = fake_monitor_unit

        from ubo_app.store.services.ethernet import NetState
        from ubo_app.utils import server

        server.send_command = lambda command, *_, has_output=False: Fake(
            _Fake__await_value={
                'connection': Fake(items=[NetState.CONNECTED]),
                'users': Fake(items=['username:password']),
            }.get(command, 'done')
            if has_output
            else 0,  # python-fake will ignore `await_value` if it is `None`
        )

        from ubo_app.utils import network

        network.has_gateway = Fake(
            _Fake__return_value=Fake(_Fake__await_value=True),
        )

    import ubo_app.display as _  # noqa: F401

    if not IS_TEST_ENV:
        signal.signal(signal.SIGTERM, signal_handler)
        signal.signal(signal.SIGINT, signal_handler)

    from ubo_gui import setup as setup_ubo_gui

    setup_ubo_gui()

    from ubo_app.service import start_event_loop_thread

    start_event_loop_thread(asyncio.new_event_loop())

    from ubo_app.store.ubo_actions import register_application
    from ubo_app.utils.gui import RawImageViewer, RawTextViewer

    register_application(
        application=RawTextViewer,
        application_id='ubo:raw-text-viewer',
    )
    register_application(
        application=RawImageViewer,
        application_id='ubo:raw-image-viewer',
    )
    import ubo_app.store.main  # noqa: F401


def clear_signal_handlers() -> None:
    """Clear the signal handlers."""
    from kivy.clock import mainthread

    mainthread(_clear_signal_handlers)()


def _clear_signal_handlers() -> None:
    signal.signal(signal.SIGTERM, signal.SIG_DFL)
    signal.signal(signal.SIGINT, signal.SIG_DFL)
    signal.signal(signal.SIGUSR1, signal.SIG_DFL)


def signal_handler(signum: int, _: object) -> None:
    """Handle the signal."""
    from ubo_app.logger import logger
    from ubo_app.utils.error_handlers import get_all_thread_stacks_string

    logger.debug(get_all_thread_stacks_string())

    from ubo_app.display import display

    _clear_signal_handlers()

    logger.info('Received signal %s, turning off the display...', signum)
    display.turn_off()

    if signum == signal.SIGINT:
        logger.info('Exiting gracefully, sending the signal again will force exit!')
        from ubo_app.store.main import store

        store.dispatch(FinishAction())
    elif signum == signal.SIGTERM:
        logger.info(
            'Exiting forcefully, sending the signal again will not be caught!',
        )
        import os

        os.kill(os.getpid(), signal.SIGTERM)



"""Reducer for the settings state."""

from __future__ import annotations

from dataclasses import replace

from redux import (
    CompleteReducerResult,
    InitAction,
    InitializationActionError,
    ReducerResult,
)
from ubo_gui.constants import SUCCESS_COLOR

from ubo_app.store.services.notifications import (
    Importance,
    Notification,
    NotificationDisplayType,
    NotificationsAddAction,
)
from ubo_app.store.settings.types import (
    ServicesStatus,
    SettingsAction,
    SettingsEvent,
    SettingsReportServiceErrorAction,
    SettingsServiceSetIsEnabledAction,
    SettingsServiceSetLogLevelAction,
    SettingsServiceSetShouldRestartAction,
    SettingsServiceSetStatusAction,
    SettingsSetServicesAction,
    SettingsStartServiceAction,
    SettingsStartServiceEvent,
    SettingsState,
    SettingsStopServiceAction,
    SettingsStopServiceEvent,
    SettingsToggleBetaVersionsAction,
    SettingsTogglePdbSignalAction,
    SettingsToggleVisualDebugAction,
)
from ubo_app.store.update_manager.types import UpdateManagerRequestCheckAction


def reducer(
    state: SettingsState | None,
    action: SettingsAction | InitAction,
) -> ReducerResult[
    SettingsState,
    UpdateManagerRequestCheckAction | NotificationsAddAction,
    SettingsEvent,
]:
    """Reducer for the settings state."""
    if state is None:
        if isinstance(action, InitAction):
            return SettingsState()

        raise InitializationActionError(action)

    match action:
        case SettingsTogglePdbSignalAction():
            from ubo_app.store.services.speech_synthesis import ReadableInformation

            return CompleteReducerResult(
                state=replace(
                    state,
                    pdb_signal=not state.pdb_signal,
                ),
                actions=[
                    NotificationsAddAction(
                        notification=Notification(
                            title='PDB Debug',
                            content='Instructions',
                            extra_information=ReadableInformation(
                                text='First make sure ipdb is installed by running:\n\n'
                                '/opt/ubo/env/bin/pip install ipdb\n\n'
                                'You need to run it only once.\n'
                                'Then send a SIGUSR1 signal to the process:\n\n'
                                'kill -SIGUSR1 <PID>',
                                picovoice_text='',
                                piper_text='',
                            ),
                            icon='',
                            importance=Importance.MEDIUM,
                            color=SUCCESS_COLOR,
                            display_type=NotificationDisplayType.STICKY,
                        ),
                    ),
                ]
                if not state.pdb_signal
                else [],
            )

        case SettingsToggleVisualDebugAction():
            return replace(
                state,
                visual_debug=not state.visual_debug,
            )

        case SettingsToggleBetaVersionsAction():
            return CompleteReducerResult(
                state=replace(
                    state,
                    beta_versions=not state.beta_versions,
                ),
                actions=[
                    NotificationsAddAction(
                        notification=Notification(
                            title='Beta Versions',
                            content='Go to the About menu and select "Recent Versions" '
                            'to continue.',
                            icon='󰒓',
                            color=SUCCESS_COLOR,
                        ),
                    ),
                    UpdateManagerRequestCheckAction(),
                ]
                if not state.beta_versions
                else [],
            )

        case SettingsSetServicesAction():
            enabled_services = [
                service for service in action.services.values() if service.is_enabled
            ]
            return CompleteReducerResult(
                state=replace(state, services=action.services),
                events=[
                    SettingsStartServiceEvent(
                        service_id=service.id,
                        delay=index * action.gap_duration,
                    )
                    for index, service in enumerate(enabled_services)
                ],
            )

        case SettingsStartServiceAction():
            return CompleteReducerResult(
                state=state,
                events=[SettingsStartServiceEvent(service_id=action.service_id)],
            )

        case SettingsStopServiceAction():
            return CompleteReducerResult(
                state=state,
                events=[SettingsStopServiceEvent(service_id=action.service_id)],
            )

        case SettingsServiceSetStatusAction() if state.services:
            service = state.services.get(action.service_id)
            if service:
                events: list[SettingsEvent] = []
                if (
                    not action.is_active
                    and service.is_active
                    and service.should_auto_restart
                ):
                    events = [
                        SettingsStartServiceEvent(
                            service_id=action.service_id,
                            delay=2,
                        ),
                    ]
                new_services = {
                    **state.services,
                    action.service_id: replace(
                        state.services[action.service_id],
                        is_active=action.is_active,
                    ),
                }
                all_services_active = all(
                    service.is_active for service in new_services.values()
                )
                return CompleteReducerResult(
                    state=replace(
                        state,
                        services=new_services,
                        services_status=ServicesStatus.READY
                        if all_services_active
                        else ServicesStatus.LOADING,
                    ),
                    events=events,
                )

            return state

        case SettingsReportServiceErrorAction():
            return replace(
                state,
                services={
                    key: replace(value, errors=[*value.errors, action.error])
                    if key == action.service_id
                    else value
                    for key, value in state.services.items()
                }
                if state.services
                else {},
            )

        case SettingsServiceSetIsEnabledAction():
            return replace(
                state,
                services={
                    **state.services,
                    action.service_id: replace(
                        state.services[action.service_id],
                        is_enabled=action.is_enabled,
                    ),
                }
                if state.services
                else {},
            )

        case SettingsServiceSetLogLevelAction():
            return replace(
                state,
                services={
                    **state.services,
                    action.service_id: replace(
                        state.services[action.service_id],
                        log_level=action.log_level,
                    ),
                }
                if state.services
                else {},
            )

        case SettingsServiceSetShouldRestartAction():
            return replace(
                state,
                services={
                    **state.services,
                    action.service_id: replace(
                        state.services[action.service_id],
                        should_auto_restart=action.should_auto_restart,
                    ),
                }
                if state.services
                else {},
            )

        case _:
            return state
